# -*- coding: utf-8 -*-
import base64, binascii, json, os, re, mimetypes, uuid, zlib
from io import StringIO, TextIOBase, BytesIO, BufferedRandom, IOBase, BufferedReader
from collections import deque
from urllib.parse import parse_qsl as parse_query_string_list, urlencode, urlencode as url_encode

from .utils import imultidict, multidict
from .ios import AsyncIO

from .hdrs import CONTENT_DISPOSITION, CONTENT_ENCODING, CONTENT_LENGTH, CONTENT_TRANSFER_ENCODING, CONTENT_TYPE
from .helpers import content_disposition_header, CHAR, TOKEN
from .protocol import ZLIB_COMPRESSOR, BROTLI_COMPRESSOR
from .exceptions import ContentEncodingError

BIG_CHUNK_LIMIT = 1<<16
DEFAULT_CONTENT_TYPE = 'application/octet-stream'
    
VALID_TCHAR_RP = re.compile(br'\A[!#$%&\'*+\-.^_`|~\w]+\Z')
INVALID_QDTEXT_CHAR_RP = re.compile(br'[\x00-\x08\x0A-\x1F\x7F]')

def create_payload(data, kwargs):
    """
    Creates a new ``PayloadBase`` instance with the given parameters.
    
    Parameters
    ----------
    data : (`list` of ``PayloadBase`` instances), ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `str`, \
            `BytesIO`, `StringIO`, `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, \
            `async-iterable` instance
    kwargs : `dict` of (`str`, (`None` or `str`)) items
        Keyword arguments for the payload.
    
    Returns
    -------
    payload : ``PayloadBase``
    
    Raises
    ------
    LookupError
        `payload` is not serializable.
    """
    data_type = data.__class__
    if issubclass(data_type, BodyPartReader):
        type_ = BodyPartReaderPayload
    elif issubclass(data_type, (bytes, bytearray, memoryview)):
        type_ = BytesPayload
    elif issubclass(data_type, str):
        type_ = StringPayload
    elif issubclass(data_type, BytesIO):
        type_ = BytesIOPayload
    elif issubclass(data_type, StringIO):
        type_ = StringIOPayload
    elif issubclass(data_type, TextIOBase):
        type_ = TextIOPayload
    elif issubclass(data_type, (BufferedReader, BufferedRandom)):
        type_ = BufferedReaderPayload
    elif issubclass(data_type, IOBase):
        type_ = IOBasePayload
    elif issubclass(data_type, AsyncIO):
        type_ = AsyncIOPayload
    elif hasattr(data_type, '__aiter__'):
        type_ = AsyncIterablePayload
    else:
        raise LookupError(data_type)
    
    return type_(data, kwargs)


class PayloadBase(object):
    """
    Base class for payloads.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : (`list` of ``PayloadBase`` instances), ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `str`, \
            `BytesIO`, `StringIO`, `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, \
            `async-iterable` instance
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : `imultidict` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ('content_type', 'data', 'encoding', 'filename', 'headers', 'size', )
    def __init__(self, data, kwargs):
        """
        Creates a new ``PayloadBase`` instance.
        
        Parameters
        ----------
        data : (`list` of ``PayloadBase`` instances), ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `str`, \
                `BytesIO`, `StringIO`, `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, \
                `async-iterable` instance
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        self.data = data
        self.encoding = kwargs.get('encoding')
        self.filename = filename = kwargs.get('filename')
        self.size = None
        
        headers = kwargs.get('headers')
        if (headers is None) or (not headers):
            headers = imultidict()
        else:
            headers = imultidict(headers)
            if ('content_type' not in kwargs) and (CONTENT_TYPE in headers):
                content_type = headers[CONTENT_TYPE]
        
        self.headers = headers
        
        content_type = kwargs.get('content_type')
        if content_type is None:
            if filename is None:
                content_type = DEFAULT_CONTENT_TYPE
            else:
                mime = mimetypes.guess_type(self.filename)[0]
                if mime is None:
                    content_type = DEFAULT_CONTENT_TYPE
                else:
                    content_type = mime
        
        self.content_type = content_type
    
    def set_content_disposition(self, disposition_type, params, quote_fields=True):
        """
        Sets content disposition header to the payload.
        
        Parameters
        ----------
        disposition_type : `str`
            The disposition's type.
        params : `dict` of (`str`, `str`) items
            Additional parameters.
        quote_fields : `bool`
            Whether field values should be quoted.
        """
        headers = self.headers
        headers.popall(CONTENT_DISPOSITION, None)
        headers[CONTENT_DISPOSITION] = content_disposition_header(disposition_type, params, quote_fields=quote_fields)
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        pass

class BytesPayload(PayloadBase):
    """
    Payload class for `bytes-like` objects.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `bytes`, `bytearray`, `memoryview` instance
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``BytesPayload`` instance.
        
        Parameters
        ----------
        data : `bytes`, `bytearray`, `memoryview` instance
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        kwargs.setdefault('content_type', DEFAULT_CONTENT_TYPE)
        
        PayloadBase.__init__(self, data, kwargs)
        
        self.size = len(data)
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        await writer.write(self.data)


class StringPayload(BytesPayload):
    """
    Payload class for `bytes-like` objects.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `bytes`
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    def __init__(self, data, kwargs):
        """
        Creates a new ``StringPayload`` instance.
        
        Parameters
        ----------
        data : `str` instance
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        encoding = kwargs.get('encoding')
        content_type = kwargs.get('content_type')
        if encoding is None:
            if content_type is None:
                encoding = 'utf-8'
                content_type = 'text/plain; charset=utf-8'
                
                kwargs['content_type'] = content_type
            else:
                mimetype = MimeType(content_type)
                encoding = mimetype.params.get('charset', 'utf-8')
            
            kwargs['encoding'] = encoding
        
        else:
            if content_type is None:
                content_type = f'text/plain; charset={encoding}'
                kwargs['content_type'] = content_type
        
        data = data.encode(encoding)
        
        BytesPayload.__init__(self, data, kwargs)


class StringIOPayload(StringPayload):
    """
    Payload class for string io objects.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `bytes`
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    def __init__(self, data, kwargs):
        """
        Creates a new ``StringIOPayload`` instance.
        
        Parameters
        ----------
        data : `StringIO` instance
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        data = data.read()
        StringPayload.__init__(self, data, kwargs)


class IOBasePayload(PayloadBase):
    """
    Payload class for `IOBase` instances.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `IOBase` instance
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    def __init__(self, data, kwargs):
        """
        Creates a new ``IOBasePayload`` instance.
        
        Parameters
        ----------
        data : `IOBase` instance
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        if 'filename' not in kwargs:
            kwargs['filename'] = getattr(data, 'name', None)
        
        PayloadBase.__init__(self, data, kwargs)
        
        try:
            disposition  = kwargs['disposition']
        except KeyError:
            disposition = 'attachment'
        else:
            disposition = None
        
        if (disposition is not None):
            filename = self.filename
            if (filename is not None):
                self.set_content_disposition(disposition, {'filename': filename})
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        data = self.data
        try:
            while True:
                chunk = data.read(BIG_CHUNK_LIMIT)
                if chunk:
                    await writer.write(chunk)
                else:
                    break
        finally:
            data.close()


class TextIOPayload(IOBasePayload):
    """
    Payload class for `TextIOBase` instances.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `TextIOBase` instance
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``TextIOPayload`` instance.
        
        Parameters
        ----------
        data : `TextIOBase` instance
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        encoding = kwargs.get('encoding')
        content_type = kwargs.get('content_type')
        if encoding is None:
            if content_type is None:
                encoding = 'utf-8'
                content_type = 'text/plain; charset=utf-8'
                
                kwargs['content_type'] = content_type
            else:
                mimetype = MimeType(content_type)
                encoding = mimetype.params.get('charset', 'utf-8')
            
            kwargs['encoding'] = encoding
        else:
            if content_type is None:
                content_type = f'text/plain; charset={encoding}'
                kwargs['content_type'] = content_type
        
        IOBasePayload.__init__(self, data, kwargs)
        
        try:
            size = os.fstat(data.fileno()).st_size - data.tell()
        except OSError:
            # `data.fileno()` is not supported. Example: `io.BufferedReader(io.BytesIO(b'data'))`
            size = None
        
        self.size = size
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        data = self.data
        try:
            while True:
                chunk = data.read(BIG_CHUNK_LIMIT)
                if chunk:
                    await writer.write(chunk.encode(self.encoding))
                else:
                    break
        finally:
            data.close()


class BytesIOPayload(IOBasePayload):
    """
    Payload class for `BytesIO` instances.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `BytesIO` instance
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    def __init__(self, data, kwargs):
        """
        Creates a new ``BytesIOPayload`` instance.
        
        Parameters
        ----------
        data : `BytesIO` instance
            The payload's data.
        kwargs : ``imultidict`` of (`str`, `str`) items
            Additional keyword arguments.
        """
        IOBasePayload.__init__(self, data, kwargs)
        
        position = data.tell()
        end = data.seek(0, os.SEEK_END)
        data.seek(position)
        self.size = end-position


class BufferedReaderPayload(IOBasePayload):
    """
    Payload class for `BufferedReader` and for `BufferedRandom` instances.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `BufferedReader`, `BufferedRandom` instance
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    def __init__(self, data, kwargs):
        """
        Creates a new ``BufferedReaderPayload`` instance.
        
        Parameters
        ----------
        data : `BufferedReader`, `BufferedRandom` instance
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        IOBasePayload.__init__(self, data, kwargs)
        try:
            size = os.fstat(data.fileno()).st_size - data.tell()
        except OSError:
            # `data.fileno()` is not supported. Example: `io.BufferedReader(io.BytesIO(b'data'))`
            size = None
        
        self.size = size


class JsonPayload(BytesPayload):
    """
    Payload class for `json` data.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data :`bytes`
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ()
    def __init__(self, data, kwargs):
        """
        Creates a new ``AsyncIterablePayload`` instance.
        
        Parameters
        ----------
        data : `None`, `str`, `int`, `float`, `list` of repeat, `dict` of (`str`, repeat) items
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        encoding = kwargs.get('encoding')
        if (encoding is None):
            kwargs['encoding'] = encoding = 'utf-8'
        
        data = json.dumps(data).encode(encoding)
        
        kwargs.setdefault('content_type', 'application/json')
        BytesPayload.__init__(self, data, kwargs)


class AsyncIterablePayload(PayloadBase):
    """
    Payload class for `async-iterable`-s.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data :`async-iterable`
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    __slots__ = ('_iterator')
    
    def __init__(self, data, kwargs):
        """
        Creates a new ``AsyncIterablePayload`` instance.
        
        Parameters
        ----------
        data : `async-iterable`
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        kwargs.setdefault('content_type', DEFAULT_CONTENT_TYPE)
        
        PayloadBase.__init__(self, data, kwargs)
        self._iterator = data.__class__.__aiter__(data)
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        iterator = self._iterator
        anext = iterator.__class__.__anext__
        try:
            while True:
                chunk = await anext(iterator)
                await writer.write(chunk)
        except StopAsyncIteration:
            self._iterator = None


class AsyncIOPayload(IOBasePayload):
    """
    Payload class for ``AsyncIO`` instances.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data :`async-iterable`
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        data = self.data
        try:
            while True:
                chunk = await data.read(BIG_CHUNK_LIMIT)
                await writer.write(chunk)
                if len(chunk) < BIG_CHUNK_LIMIT:
                    break
        finally:
            data.close()


class BodyPartReaderPayload(PayloadBase):
    """
    Payload class for ``BodyPartReader``.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data :``BodyPartReader``
        The payload itself.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    """
    def __init__(self, data, kwargs):
        """
        Creates a new ``BodyPartReaderPayload`` instance.
        
        Parameters
        ----------
        data : ``BodyPartReader``
            The payload's data.
        kwargs : `dict` of (`str`, `Any`) items
            Additional keyword arguments.
        """
        PayloadBase.__init__(self, data, kwargs)
        
        params = {}
        name = data.name
        if (name is not None):
            params['name'] = name
        
        filename = data.filename
        if (filename is not None):
            params['filename'] = filename
        
        if params:
            self.set_content_disposition('attachment', params)
    
    async def write(self, writer):
        """
        Writes the payload to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        """
        field = self.data
        while True:
            chunk = await field.read_chunk(size=65536)
            if chunk:
                await writer.write(field.decode(chunk))
            else:
                break

class MimeType(object):
    # Parses a MIME type into its components
    
    __slots__ = ('mtype', 'params', 'stype', 'suffix',)
    def __init__(self, mimetype):
        
        if not mimetype:
            self.mtype = ''
            self.stype = ''
            self.suffix = ''
            self.params = {}
            return
        
        parts = mimetype.split(';')
        params = multidict()
        for item in parts[1:]:
            if not item:
                continue
            if '=' in item:
                key, value = item.split('=', 1)
            else:
                key = item
                value = ''
            
            params[key.strip().lower()] = value.strip(' "')
        
        
        fulltype = parts[0].strip().lower()
        if fulltype == '*':
            fulltype = '*/*'
        
        if '/' in fulltype:
            mtype, stype = fulltype.split('/', 1)
        else:
            mtype = fulltype
            stype = ''

        if '+' in stype:
            stype, suffix = stype.split('+', 1)
        else:
            suffix = ''
            
        self.mtype = mtype
        self.stype = stype
        self.suffix = suffix
        self.params = params
    
    def __repr__(self):
        return (f'<{self.__class__.__name__} mtype={self.mtype!r} stype={self.stype!r} suffix={self.suffix!r} params='
            f'{self.params!r}>')

    __str__ = __repr__


def parse_content_disposition(header):
    def is_token(string):
        return string and TOKEN >= set(string)

    def is_quoted(string):
        return string[0] == string[-1] == '"'

    def is_rfc5987(string):
        return is_token(string) and string.count("'") == 2

    def is_extended_param(string):
        return string.endswith('*')

    def is_continuous_param(string):
        pos = string.find('*') + 1
        if not pos:
            return False
        substring = string[pos:-1] if string.endswith('*') else string[pos:]
        return substring.isdigit()

    def unescape(text, *, chars=''.join(map(re.escape, CHAR))):
        return re.sub(f'\\\\([{chars}])', '\\1', text)

    if not header:
        return None, {}

    disposition_type, *parts = header.split(';')
    if not is_token(disposition_type):
        return None, {}

    params = {}
    while parts:
        item = parts.pop(0)
        
        if '=' not in item:
            return None, {}
        
        key, value = item.split('=', 1)
        key = key.lower().strip()
        value = value.lstrip()
        
        if key in params:
            return None, {}
        
        if not is_token(key):
            continue
        
        elif is_continuous_param(key):
            if is_quoted(value):
                value = unescape(value[1:-1])
            elif not is_token(value):
                continue
        
        elif is_extended_param(key):
            if is_rfc5987(value):
                encoding, _, value = value.split("'", 2)
                encoding = encoding or 'utf-8'
            else:
                continue
            
            try:
                value = unquote(value, encoding, 'strict')
            except UnicodeDecodeError:
                continue
        
        else:
            failed = True
            
            if is_quoted(value):
                value = unescape(value[1:-1].lstrip('\\/'))
            elif is_token(value):
                failed = False
            elif parts:
                value_ = f'{value};{parts[0]}'
                if is_quoted(value_):
                    parts.pop(0)
                    value = unescape(value_[1:-1].lstrip('\\/'))
                    failed = False
            
            if failed:
                return None, {}
        
        params[key] = value
    
    return disposition_type.lower(), params


def content_disposition_filename(params,name='filename'):
    name_suf = f'{name}*'
    if not params:
        return None
    elif name_suf in params:
        return params[name_suf]
    elif name in params:
        return params[name]
    else:
        parts = []
        fnparams = sorted((key, value) for key, value in params.items() if key.startswith('filename*'))
        for num, (key, value) in enumerate(fnparams):
            _, tail = key.split('*', 1)
            if tail.endswith('*'):
                tail = tail[:-1]
            if tail == str(num):
                parts.append(value)
            else:
                break
        if not parts:
            return None
        value = ''.join(parts)
        if "'" in value:
            encoding, _, value = value.split("'", 2)
            encoding = encoding or 'utf-8'
            return unquote(value, encoding, 'strict')
        return value

# TODO
class BodyPartReader(object):
    # Multipart reader for single body part.
    __slots__ = ('headers', '_boundary', '_content', '_at_eof', '_length', '_read_bytes', '_unread', '_prev_chunk',
        '_content_eof', '_cache',)
    
    chunk_size = 8192
    
    def __init__(self, boundary, headers, content):
        self.headers = headers
        self._boundary = boundary
        self._content = content
        self._at_eof = False
        length = self.headers.get(CONTENT_LENGTH)
        self._length = None  if length is None else int(length)
        self._read_bytes = 0
        self._unread = deque()
        self._prev_chunk = None
        self._content_eof = 0
        self._cache = {}
        
    def __aiter__(self):
        return self

    async def __anext__(self):
        item = await self.read()
        if item:
            return item
        raise StopAsyncIteration

    async def next(self):
        item = await self.read()
        if not item:
            return None
        return item

    async def read(self, *, decode=False):
        # Reads body part data.
        #decode : `bool`, Optional : should decode data following by encoding method from `Content-Encoding` header.
        #     Defaults to `False`.
        if self._at_eof:
            return b''
        data = bytearray()
        while not self._at_eof:
            data.extend((await self.read_chunk(self.chunk_size)))
        if decode:
            return self.decode(data)
        return data
    
    async def read_chunk(self, size=chunk_size):
        # Reads body part content chunk of the specified size.
        if self._at_eof:
            return b''
        if self._length:
            chunk = await self._read_chunk_from_length(size)
        else:
            chunk = await self._read_chunk_from_stream(size)
        
        self._read_bytes += len(chunk)
        if self._read_bytes == self._length:
            self._at_eof = True
        if self._at_eof:
            await self._content.readline()
        
        return chunk

    async def _read_chunk_from_length(self, size):
        # Reads body part content chunk of the specified size.
        chunk_size = min(size,self._length-self._read_bytes)
        chunk = await self._content.read(chunk_size)
        return chunk
    
    async def _read_chunk_from_stream(self, size):
        # Reads content chunk of body part with unknown length.
        first_chunk = self._prev_chunk is None
        if first_chunk:
            self._prev_chunk = await self._content.read(size)
        
        chunk = await self._content.read(size)
        self._content_eof += int(self._content.at_eof())
        
        window = self._prev_chunk + chunk
        sub = b'\r\n' + self._boundary
        
        if first_chunk:
            idx = window.find(sub)
        else:
            idx = window.find(sub, max(0, len(self._prev_chunk)-len(sub)))
        if idx >= 0:
            # pushing boundary back to content
            self._content.unread_data(window[idx:])
            if size > idx:
                self._prev_chunk = self._prev_chunk[:idx]
            chunk = window[len(self._prev_chunk):idx]
            if not chunk:
                self._at_eof = True
        result = self._prev_chunk
        self._prev_chunk = chunk
        return result
    
    async def readline(self):
        if self._at_eof:
            return b''
        
        if self._unread:
            line = self._unread.popleft()
        else:
            line = await self._content.readline()
        
        if line.startswith(self._boundary):
            # the very last boundary may not come with \r\n,
            # so set single rules for everyone
            sline = line.rstrip(b'\r\n')
            boundary = self._boundary
            last_boundary = self._boundary + b'--'
            #make sure that we read exactly the boundary, not something alike
            if sline == boundary or sline == last_boundary:
                self._at_eof = True
                self._unread.append(line)
                return b''
        else:
            next_line = await self._content.readline()
            if next_line.startswith(self._boundary):
                line = line[:-2]  # strip CRLF but only once
            self._unread.append(next_line)
        
        return line

    async def release(self):
        # Like `.read`, but reads all the data.
        if self._at_eof:
            return
        
        while not self._at_eof:
            await self.read_chunk(self.chunk_size)
    
    async def text(self, *, encoding=None):
        # Like `.read`, but assumes that body part contains text data.
        # encoding : `str`, Optional : Custom text encoding. Overrides specified in charset param of `Content-Type`
        #     header
        
        data = await self.read(decode=True)
        # see https://www.w3.org/TR/html5/forms.html#multipart/form-data-encoding-algorithm
        # and https://dvcs.w3.org/hg/xhr/raw-file/tip/Overview.html#dom-xmlhttprequest-send
        encoding = encoding or self.get_charset(default='utf-8')
        return data.decode(encoding)

    async def json(self, *, encoding=None):
        # Like `.read`, but assumes that body parts contains JSON data.
        # encoding :`str`, Optional : Custom JSON encoding. Overrides specified in charset param of `Content-Type`
        #     header
        
        data = await self.read(decode=True)
        if not data:
            return None
        encoding = encoding or self.get_charset(default='utf-8')
        return json.loads(data.decode(encoding))
    
    async def form(self, *, encoding=None):
        # Like `.read`, but assumes that body parts contains form url_encoded data.
        # encoding : `str`, Optional : Custom form encoding. Overrides specified in charset param of `Content-Type`
        #     header
        data = await self.read(decode=True)
        if not data:
            return None
        encoding = encoding or self.get_charset(default='utf-8')
        return parse_query_string_list(data.rstrip().decode(encoding), keep_blank_values=True, encoding=encoding)

    def at_eof(self):
        #Returns True if the boundary was reached or False otherwise.

        return self._at_eof

    def decode(self, data):
        #Decodes data according the specified Content-Encoding
        #or Content-Transfer-Encoding headers value.
        
        #arguments: data=bytearray
        #raises RuntimeError if encoding is unknown.
        #returns bytes

        if CONTENT_TRANSFER_ENCODING in self.headers:
            data = self._decode_content_transfer(data)
        if CONTENT_ENCODING in self.headers:
            return self._decode_content(data)
        return data

    def _decode_content(self, data):
        encoding = self.headers[CONTENT_ENCODING].lower()

        if encoding == 'deflate':
            return zlib.decompress(data, -zlib.MAX_WBITS)
        elif encoding == 'gzip':
            return zlib.decompress(data, 16 + zlib.MAX_WBITS)
        elif encoding == 'identity':
            return data
        else:
            raise RuntimeError(f'unknown content encoding: {encoding}')

    def _decode_content_transfer(self, data):
        encoding = self.headers[CONTENT_TRANSFER_ENCODING].lower()

        if encoding == 'base64':
            return base64.b64decode(data)
        elif encoding == 'quoted-printable':
            return binascii.a2b_qp(data)
        elif encoding in ('binary', '8bit', '7bit'):
            return data
        else:
            raise RuntimeError(f'unknown content transfer encoding: {encoding}')

    def get_charset(self, default=None):
        #Returns charset parameter from Content-Type header or default.

        ctype = self.headers.get(CONTENT_TYPE, '')
        mimetype = MimeType(ctype)
        return mimetype.params.get('charset', default)

    @property
    def name(self):
        #Returns name specified in Content-Disposition header or None
        #if missed or header is malformed.
        _, params = parse_content_disposition(self.headers.get(CONTENT_DISPOSITION))
        return content_disposition_filename(params, 'name')
    
    @property
    def filename(self):
        #Returns filename specified in Content-Disposition header or None
        #if missed or header is malformed
        _, params = parse_content_disposition(self.headers.get(CONTENT_DISPOSITION))
        return content_disposition_filename(params, 'filename')


class MultipartWriter(PayloadBase):
    """
    Multipart body writer.
    
    Attributes
    ----------
    content_type : `None` or `str`
        The payload's content type.
    data : `list` of ``PayloadBase`` instances
        The contained payloads.
    filename : `None` or `str`
        The payload's file's name if applicable.
    encoding : `None` or`str`
        Encoding used to encode the payload's data.
    headers : ``imultidict`` of (`str`, `str`) items
        Payload specific headers.
    size : `None` or `int`
        The payload's size if applicable.
    _boundary : `bytes`
        Boundary to matk the payload's start and end.
    """
    __slots__ = ('_boundary', )
    
    def __init__(self, subtype='mixed', boundary=None):
        """
        Creates a new ``MultipartWriter`` instance with the given parameters.
        
        Parameters
        ----------
        subtype : `str`, Optional
            The subtype of the mutlipart writer. Defaults to `'mixed'`, but also can `'form-data'` for example.
        boundary : `Nome` or `str`, Optional
            Boundary to matk the payload's start and end. If not given or gvien as `None`, then is autogenerated.
        
        Raises
        ------
        UnicodeEncodeError
            `boundary` is given as `str`, but is not `ascii` encodable.
        ValueError
            `boundary` contains invalid character.
        """
        if (boundary is None):
            boundary = uuid.uuid4().hex.encode('ascii')
        else:
            try:
                boundary = boundary.encode('ascii')
            except UnicodeEncodeError as err:
                raise ValueError('boundary should contains ASCII only chars') from err
        
        # Refer to RFCs 7231, 7230, 5234.
        #
        # parameter      = token "=" ( token / quoted-string )
        # token          = 1*tchar
        # quoted-string  = DQUOTE *( qdtext / quoted-pair ) DQUOTE
        # qdtext         = HTAB / SP / %x21 / %x23-5B / %x5D-7E / obs-text
        # obs-text       = %x80-FF
        # quoted-pair    = "\" ( HTAB / SP / VCHAR / obs-text )
        # tchar          = "!" / "#" / "$" / "%" / "&" / "'" / "*"
        #                  / "+" / "-" / "." / "^" / "_" / "`" / "|" / "~"
        #                  / DIGIT / ALPHA
        #                  ; any VCHAR, except delimiters
        # VCHAR          = %x21-7E
        
        if VALID_TCHAR_RP.match(boundary) is None:
            if INVALID_QDTEXT_CHAR_RP.search(boundary) is not None:
                raise ValueError('Boundary value contains invalid characters.')
            
            # escape %x5C and %x22
            quoted_boundary = boundary.replace(b'\\', b'\\\\')
            quoted_boundary = quoted_boundary.replace(b'"', b'\\"')
            quoted_boundary = quoted_boundary.decode('ascii')
            quoted_boundary = f'"{quoted_boundary}"'
        else:
            quoted_boundary = boundary.decode('ascii')
        
        kwargs = {'content_type': f'multipart/{subtype}; boundary={quoted_boundary}'}
        
        PayloadBase.__init__(self, [], kwargs)
        
        self._boundary = boundary
        self.headers[CONTENT_TYPE] = self.content_type
    
    @property
    def boundary(self):
        """
        Returns the multipart writer's boundary as string.
        
        Returns
        -------
        boundary : `str`
        """
        return self._boundary.decode('ascii')
    
    def append(self, body_part, headers=None):
        """
        Adds a new body part to the ``MultipartWriter``.
        
        Parameters
        ----------
        body_part : ``PayloadBase``, ``BodyPartReader``, `bytes`, `bytearray`, `memoryview`, `BytesIO`, `StringIO`, \
            `TextIOBase`, `BufferedReader`, `BufferedRandom`, `IOBase`, ``AsyncIO``, `async-iterable` instance
        headers : `None` or ``imultidict`` of (`str`, `str`) items, Optional
            Optional headers for the field.
        
        Returns
        -------
        payload : ``PayloadBase``
            The created payload.
        
        Raises
        ------
        TypeError
            Cannot create payload from the givne `body_part`.
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        if headers is None:
            headers = imultidict()
        
        if isinstance(body_part, PayloadBase):
            if (headers is not None):
                body_part.headers.update(headers)
            
            payload = body_part
        else:
            kwargs = {}
            if (headers is not None):
                kwargs['headers'] = headers
            
            try:
                payload = create_payload(body_part, kwargs)
            except LookupError as err:
                raise TypeError(f'Cannot create payload from: {body_part!r}') from err
        
        self.append_payload(payload)
        return payload
    
    def append_payload(self, payload):
        """
        Adds a payload to the multipart writer.
        
        Parameters
        ----------
        payload : ``PayloadBase`` instance
            The payload to add to the body.
        
        Raises
        ------
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        # content-type
        payload_headers = payload.headers
        if CONTENT_TYPE not in payload_headers:
            payload_headers[CONTENT_TYPE] = payload.content_type
        
        # content-encoding or compression
        try:
            content_encoding = payload_headers[CONTENT_ENCODING].lower()
        except KeyError:
            content_encoding = None
        else:
            if content_encoding in ('deflate', 'gzip', 'br', ):
                pass
            elif content_encoding in ('', 'identity'):
                content_encoding = None
            else:
                raise RuntimeError(f'Unknown content-encoding: {content_encoding!r}.')
        
        # te-encoding
        try:
            transfer_encoding = payload_headers[CONTENT_TRANSFER_ENCODING].lower()
        except KeyError:
            transfer_encoding = None
        else:
            if transfer_encoding == '':
                transfer_encoding = None
            elif transfer_encoding in ('base64', 'quoted-printable'):
                pass
            elif transfer_encoding == 'binary':
                transfer_encoding = None
            else:
                raise RuntimeError(f'Unknown content transfer encoding: {transfer_encoding!r}.')
        
        # Set size to payload headers if applicable.
        size = payload.size
        if (size is not None) and (content_encoding is None) and (transfer_encoding is None):
            payload_headers[CONTENT_LENGTH] = str(size)
        
        # Render headers.
        result = []
        extend = result.extend
        for k, v in payload_headers.items():
            extend((k, ': ', v, '\r\n'))
        
        result.append('\r\n')
        
        headers = ''.join(result).encode('utf-8')
        
        # Calculate new size.
        data = self.data
        if data:
            size = self.size
        else:
            size = 6+len(self._boundary)
            # b'--'+self._boundary+b'--\r\n'
        
        if (size is not None):
            if (content_encoding is not None) or (transfer_encoding is not None) or (payload.size is None):
                size = None
            else:
                size += 6+len(self._boundary)+payload.size+len(headers)
                # b'--'+self._boundary+b'\r\n' ... data ... b'\r\n'
            
            self.size = size
        
        # Add part.
        data.append((payload, headers, content_encoding, transfer_encoding))
    
    
    def append_json(self, obj, headers=None):
        """
        Helper method to add a json field.
        
        Parameters
        ----------
        obj : `None`, `str`, `int`, `float`, `list` of repeat, `dict` of (`str`, repeat) items
            The payload's data.
        headers : `None` or ``imultidict`` of (`str`, `str`) items, Optional
            Optional headers for the json field.

        Returns
        -------
        payload : ``JsonPayload``
            The created json payload.
        
        Raises
        ------
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        kwargs = {}
        if (headers is not None):
            kwargs['headers'] = headers
        
        payload = JsonPayload(obj, kwargs)
        self.append_payload(payload)
        return payload
    
    
    def append_form(self, obj, headers=None):
        """
        Helper method to add url_encoded field.
        
        Parameters
        ----------
        obj : `mapping` of (`str`, `Any`) items, `sequence` of `tuple` (`str`, `Any`) items
            The object, what should be percent encoded for a post request.
        headers : `None` or ``imultidict`` of (`str`, `str`) items, Optional
            Optional headers for the url_encoded field.
        
        Returns
        -------
        payload : ``StringPayload``
            The created string payload.
        
        Raises
        ------
        RuntimeError
            - The `payload`'s content has unknown content-encoding.
            - The `payload`'s content has unknown content-transfer-encoding.
        """
        if hasattr(obj.__class__, 'items'): # mapping type
            obj = list(obj.items())
        
        data = url_encode(obj, doseq=True)
        
        kwargs = {'content_type': 'application/x-www-form-url_encoded'}
        
        if (headers is not None):
            kwargs['headers'] = headers
        
        payload = StringPayload(data, kwargs)
        self.append_payload(payload)
        return payload
    
    
    async def write(self, writer, close_boundary=True):
        """
        Writes the payloads of the multipart writer to the given http writer.
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Http writer to write the payload's data to.
        close_boundary : `bool`, Optional
            Whether the multypart's payload should be closed with it's boundary.
        """
        data = self.data
        if not data:
            return
        
        for part, headers, content_encoding, transfer_encoding in data:
            await writer.write(b'--' + self._boundary + b'\r\n') # fb strings pls!
            await writer.write(headers)
            
            if (content_encoding is not None) or (transfer_encoding is not None):
                multipart_payload_writer = MultipartPayloadWriter(writer, content_encoding, transfer_encoding)
                
                await part.write(multipart_payload_writer)
                await multipart_payload_writer.write_eof()
            else:
                await part.write(writer)
            await writer.write(b'\r\n')
        
        if close_boundary:
            await writer.write(b'--'+self._boundary+b'--\r\n')


class MultipartPayloadWriter(object):
    __slots__ = ('compressor', 'encoding_buffer', 'transfer_encoding', 'writer',)
    
    def __init__(self, writer, content_encoding, transfer_encoding):
        
        if content_encoding == 'gzip':
            compressor = ZLIB_COMPRESSOR(wbits=16+zlib.MAX_WBITS)
        elif content_encoding == 'deflate':
            compressor = ZLIB_COMPRESSOR(wbits=-zlib.MAX_WBITS)
        elif content_encoding == 'br':
            if BROTLI_COMPRESSOR is None:
                raise ContentEncodingError('Can not decode content-encoding: brotli (br). Please install `brotlipy`.')
            compressor = BROTLI_COMPRESSOR()
        elif content_encoding == 'identity':
            # I asume this is no encoding
            compressor = None
        else:
            raise ContentEncodingError(f'Can not decode content-encoding: {content_encoding!r}.')
        
        if transfer_encoding == 'base64':
            transfer_encoding = transfer_encoding
            encoding_buffer = bytearray()
        else:
            if transfer_encoding == 'quoted-printable':
                transfer_encoding = transfer_encoding
            else:
                transfer_encoding = None
            encoding_buffer = None
        
        self.writer = writer
        self.transfer_encoding = transfer_encoding
        self.compressor = compressor
        self.encoding_buffer = encoding_buffer
    
    async def write_eof(self):
        compressor = self.compressor
        if (compressor is not None):
            self.compressor = None
            chunk = compressor.flush()
                
            if chunk:
                await self.write(chunk)
        
        if self.transfer_encoding == 'base64':
            encoding_buffer = self.encoding_buffer
            if encoding_buffer:
                await self.writer.write(base64.b64encode(encoding_buffer))
    
    async def write(self, chunk):
        compressor = self.compressor
        if (compressor is not None):
            if chunk:
                chunk = compressor.compress(chunk)
                if not chunk:
                    return
        
        transfer_encoding = self.transfer_encoding
        if transfer_encoding == 'base64':
            encoding_buffer = self.encoding_buffer
            encoding_buffer.extend(chunk)
            
            if encoding_buffer:
                barrier = (len(encoding_buffer)//3)*3
                if barrier:
                    encoding_chunk = encoding_buffer[:barrier]
                    del encoding_buffer[:barrier]
                    encoding_chunk = base64.b64encode(encoding_chunk)
                    await self.writer.write(encoding_chunk)
        
        elif transfer_encoding == 'quoted-printable':
            await self.writer.write(binascii.b2a_qp(chunk))
            
        else:
            await self.writer.write(chunk)





# -*- coding: utf-8 -*-
import ssl, selectors
from collections import deque
import socket as module_socket

from .futures import Future

def _create_transport_context(server_side, server_hostname):
    if server_side:
        raise ValueError('Server side SSL needs a valid SSLContext')
    
    # Client side may pass ssl=True to use a default
    # context; in that case the ssl_context passed is None.
    # The default is secure for client connections.
    ssl_context = ssl.create_default_context()
    if (server_hostname is None) or (not server_hostname):
        ssl_context.check_hostname = False
    
    return ssl_context

if hasattr(module_socket, 'TCP_NODELAY'):
    def _set_nodelay(sock):
        if (sock.family in (module_socket.AF_INET, module_socket.AF_INET6) and
            sock.type == module_socket.SOCK_STREAM and sock.proto == module_socket.IPPROTO_TCP):
            sock.setsockopt(module_socket.IPPROTO_TCP, module_socket.TCP_NODELAY, 1)
else:
    def _set_nodelay(sock):
        pass

MAX_SIZE = 262144

UNWRAPPED = 'UNWRAPPED'
DO_HANDSHAKE = 'DO_HANDSHAKE'
WRAPPED = 'WRAPPED'
SHUTDOWN = 'SHUTDOWN'

class _SSLPipe(object):
    __slots__=('_handshake_cb', '_incoming', '_outgoing', '_shutdown_cb', 'context', 'need_ssldata', 'server_hostname',
        'server_side', 'ssl_object', 'state',)
    
    def __init__(self, context, server_side, server_hostname=None):
        self.context = context
        self.server_side = server_side
        self.server_hostname = server_hostname
        self.state = UNWRAPPED
        self._incoming = ssl.MemoryBIO()
        self._outgoing = ssl.MemoryBIO()
        self.ssl_object = None
        self.need_ssldata = False
        self._handshake_cb = None
        self._shutdown_cb = None
    
    
    @property
    def wrapped(self):
        return self.state is WRAPPED
    
    def do_handshake(self, callback=None):
        if self.state is not UNWRAPPED:
            raise RuntimeError('handshake in progress or completed')
        self.ssl_object = self.context.wrap_bio(self._incoming, self._outgoing, server_side=self.server_side,
            server_hostname=self.server_hostname)
        
        self.state = DO_HANDSHAKE
        self._handshake_cb=callback
        ssldata, _ = self.feed_ssldata(b'', only_handshake=True)
        
        return ssldata
    
    def shutdown(self, callback=None):
        state = self.state
        if state is UNWRAPPED:
            raise RuntimeError('no security layer present')
        if state is SHUTDOWN:
            raise RuntimeError('shutdown in progress')
        
        self.state = SHUTDOWN
        self._shutdown_cb = callback
        ssldata, offset = self.feed_ssldata(b'')
        return ssldata
    
    def feed_eof(self):
        self._incoming.write_eof()
        self.feed_ssldata(b'') #return is ignored
    
    def feed_ssldata(self, data, only_handshake=False):
        if self.state is UNWRAPPED:
            # If unwrapped, pass plaintext data straight through.
            if data:
                appdata = [data]
            else:
                appdata = []
            return [], appdata
    
        self.need_ssldata = False
        if data:
            self._incoming.write(data)
        
        ssldata = []
        appdata = []
        try:
            state = self.state
            if state is DO_HANDSHAKE:
                # Call do_handshake() until it doesn't raise anymore.
                self.ssl_object.do_handshake()
                self.state = state = WRAPPED
                
                handshake_cb = self._handshake_cb
                if (handshake_cb is not None):
                    handshake_cb(None)
                
                if only_handshake:
                    return ssldata, appdata
                # Handshake done: execute the wrapped block
            
            if state is WRAPPED:
                # Main state: read data from SSL until close_notify
                while True:
                    chunk = self.ssl_object.read(MAX_SIZE)
                    appdata.append(chunk)
                    if not chunk:  # close_notify
                        break
            
            elif state is SHUTDOWN:
                # Call shutdown() until it doesn't raise anymore.
                self.ssl_object.unwrap()
                self.ssl_object = None
                self.state = UNWRAPPED
                
                shutdown_cb = self._shutdown_cb
                if (shutdown_cb is not None):
                    shutdown_cb()
            
            elif state is UNWRAPPED:
                # Drain possible plaintext data after close_notify.
                appdata.append(self._incoming.read())
        
        except (ssl.SSLError, ssl.CertificateError) as err:
            err_number = getattr(err, 'errno', None)
            if err_number not in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE, ssl.SSL_ERROR_SYSCALL):
                if self.state is DO_HANDSHAKE:
                    handshake_cb = self._handshake_cb
                    if (handshake_cb is not None):
                        handshake_cb(err)
                raise
            
            self.need_ssldata = (err_number == ssl.SSL_ERROR_WANT_READ)

        # Check for record level data that needs to be sent back.
        # Happens for the initial handshake and renegotiations.
        if self._outgoing.pending:
            ssldata.append(self._outgoing.read())
        return ssldata, appdata

    def feed_appdata(self, data, offset):
        ssldata = []
        if self.state is UNWRAPPED:
            # pass through data in unwrapped mode
            if offset < len(data):
                ssldata.append(memoryview(data)[offset:])
            oofset = len(data)
        else:
            view = memoryview(data)
            while True:
                self.need_ssldata = False
                try:
                    if offset < len(view):
                        offset += self.ssl_object.write(view[offset:])
                except ssl.SSLError as err:
                    # It is not allowed to call write() after unwrap() until the close_notify is acknowledged. We return
                    # the condition to the caller as a short write.
                    if err.reason == 'PROTOCOL_IS_SHUTDOWN':
                        err.errno = ssl.SSL_ERROR_WANT_READ
                        need_ssldata = True
                    else:
                        errno = err.errno
                        if errno == ssl.SSL_ERROR_WANT_READ:
                            need_ssldata = True
                        elif errno == ssl.SSL_ERROR_WANT_WRITE or errno == ssl.SSL_ERROR_SYSCALL:
                            need_ssldata = False
                        else:
                            raise
                    
                    self.need_ssldata = need_ssldata
    
                # See if there's any record level data back for us.
                if self._outgoing.pending:
                    ssldata.append(self._outgoing.read())
                
                if offset == len(view) or self.need_ssldata:
                    break
        
        return ssldata, offset

class _SSLProtocolTransport(object):
    """
    Asynchronous ssl protocol's read-write  transport implementation.
    
    Attributes
    ----------
    app_protocol : `Any`
        Asynchronous protocol implementation.
    closed : `bool`
        Whether the ssl protocol transport is closed.
    ssl_protocol : ``SSLProtocol``
        The respective ssl protocol, what's `.app_protocol` is the ``_SSLProtocolTransport`` instance.
    """
    __slots__ = ('app_protocol', 'closed', 'ssl_protocol',)
    def __init__(self, ssl_protocol, app_protocol):
        """
        Creates a  new ``_SSLProtocolTransport`` instance.
        
        Parameters
        ----------
        ssl_protocol : ``SSLProtocol``
            The respective ssl protocol, what's `.app_protocol` is the ``_SSLProtocolTransport`` instance.
        app_protocol : `Any`
            Asynchronous protocol implementation.
        """
        self.ssl_protocol = ssl_protocol
        self.app_protocol = app_protocol
        self.closed = False
    
    def __repr__(self):
        """Returns the ssl protocol transport's representation."""
        result = [
            '<',
            self.__class__.__name__,
            ' ',
                ]
        
        if self.closed:
            result.append('closed')
            add_comma = True
        else:
            add_comma = False
        
        ssl_protocol = self.ssl_protocol
        if (ssl_protocol is not None):
            if add_comma:
                result.append(', ')
            else:
                add_comma = True
            
            result.append('ssl_protocol=')
            result.append(ssl_protocol.__class__.__name__)
        
        app_protocol = self.ssl_protocol
        if (app_protocol is not None):
            if add_comma:
                result.append(', ')
            
            result.append('app_protocol=')
            result.append(app_protocol.__class__.__name__)
        
        result.append('>')
        
        return ''.join(result)
    
    def get_extra_info(self, name, default=None):
        """
        Gets optional transport information.
        
        Parameters
        ----------
        name : `str`
            The extra information's name to get.
        default : `Any`, Optional
            Default value to return if `name` could not be macthed. Defaults to `None`.
        
        Returns
        -------
        info : `default`, `Any`
        """
        return self.ssl_protocol.get_extra_info(name, default)
    
    def set_protocol(self, protocol):
        """
        Sets a new protocol to the transport.
        
        Parameters
        ----------
        protocol : `Any`
            Asynchronous protocol implementation.
        """
        self.app_protocol = protocol
        self.ssl_protocol.app_protocol = protocol
     
    def get_protocol(self):
        """
        Gets the ssl protocol transport actual protocol.
        
        Returns
        -------
        protocol : `Any`
            Asynchronous protocol implementation.
        """
        return self.app_protocol
    
    def is_closing(self):
        """
        Returns whether the ssl protoco ltransport is closing.
        
        Returns
        -------
        is_closing : `bool`
        """
        return self.closed
    
    def close(self):
        """
        Starts the shutdown process of the ssl protocol transport.
        """
        self.closed = True
        self.ssl_protocol._start_shutdown()

    def __del__(self):
        """
        Closes the ssl protocol transport ifn ot yet clsoed.
        """
        if not self.closed:
            self.close()

    def pause_reading(self):
        """
        Pause the receiving end.
        
        No data will be passed to the respective protocol's ``.data_received`` method until ``.resume_reading`` is
        called.
        """
        transport = self.ssl_protocol.transport
        if (transport is not None):
            transport.pause_reading()

    def resume_reading(self):
        """
        Resume the receiving end.
        
        Data received will once again be passed to the respective protocol's ``.data_received`` method.
        """
        transport = self.ssl_protocol.transport
        if (transport is not None):
            transport.resume_reading()
    
    def set_write_buffer_limits(self, low=None, high=None):
        """
        Set the high- and low-water limits for write flow control.
        
        These two values control when to call the protocol's ``.pause_writing`` and ``.resume_writing`` methods. If
        specified, the low-water limit must be less than or equal to the high-water limit. Neither value can be
        negative. The defaults are implementation-specific. If only the high-water limit is given, the low-water limit
        defaults to an implementation-specific value less than or equal to the high-water limit. Setting high to zero
        forces low to zero as well, and causes ``.pause_writing`` to be called whenever the buffer becomes non-empty.
        Setting low to zero causes ``.resume_writing`` to be called only once the buffer is empty. Use of zero for
        either limit is generally sub-optimal as it reduces opportunities for doing I/O and computation concurrently.
        
        Parameters
        ----------
        high : `None` or `int`, Optional
            High limit to stop reading if reached.
        low : `None` or `int`, Optional
            Low limit to start reading if reached.
        
        Raises
        ------
        ValueError
            If `low` is lower than `0` or if `low` is higher than `high`.
        """
        transport = self.ssl_protocol.transport
        if (transport is not None):
            transport.set_write_buffer_limits(high, low)
        
    def get_write_buffer_size(self):
        """
        Return the current size of the write buffer.
        
        Returns
        -------
        get_write_buffer_size : `int`
        """
        transport = self.ssl_protocol.transport
        if transport is None:
            write_buffer_size = 0
        else:
            write_buffer_size = transport.get_write_buffer_size()
        
        return write_buffer_size
    
    def write(self, data):
        """
        Write the given `data` to the transport.
        
        Do not blocks, but queues up the data instead to be sent as can.
        
        Parameters
        ----------
        data : `bytes-like`
            The data to send.
        
        Raises
        ------
        TypeError
            If `data` was not given as `bytes-like`.
        """
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f'`data` expecting a `bytes-like`, got {data.__class__.__name__}.')
        
        if data:
            self.ssl_protocol._write_appdata(data)
    
    def writelines(self, list_of_data):
        """
        Writes an `iterable` of `bytes-like` to the transport.
        
        Parameters
        ----------
        list_of_data : `iterable` of `bytes-like`
            An iterable of data to send.
        
        Raises
        ------
        TypeError
            If `list_of_data` was not given as `iterable` of `bytes-like`.
        """
        self.write(b''.join(list_of_data))
    
    def can_write_eof(self):
        """
        Return whether the transport supports ``.write_eof``.
        
        Returns
        -------
        can_write_eof : `bool`
            ``_SSLProtocolTransport`` instances always return `False`.
        """
        return False
    
    def abort(self):
        """
        Close the transport immediately.
        
        Buffered data will be lost and no more data will be received.
        
        The respective protocol's `.connection_lost` will be eventually called with `None`.
        """
        self.ssl_protocol._abort()


class SSLProtocol(object):
    """
    Asynchornous SSL protocol implementation on top of a `socket`. Uses `ssl.MemoryBIO` instances for incoming and
    outgoing data.
    
    Attributes
    ----------
    _call_connection_made : `bool`
        Whether the the ``.app_protocol``'s `.connection_made` should be called when handshake is completed.
    _connection_made_waiter : `None` or ``Future``
        A waiter future, what's result is set when connection is made, aka handshae is completed, or if when the
        connection is lost, depending which happens first.
        
        After the future's result or exception is set, the attribute is set as `None`.
    _extra : `dict` of (`str`, `Any`) items, Optional
        Optional transport informations.
    _in_handshake : `bool`
        Whether the ssl transport is in handshaking.
    _in_shutdown : `bool`
        Whether the ssl protocol is shut or shutting shutting down.
    _session_established : `bool`
        Whether the session is established. Is set after handshake and is set back to `False` when the connection is
        lost.
    _ssl_context : `ssl.SSLContext`
        The connection's ssl type.
    _write_backlog : `deque` of `tuple` (`bytes-like`, `int`)
        Ensured data queued up to be written. Each element contains a the data to write and an offset till write from.
    app_protocol : `Any`
        Asynchronous protcol implemnetation.
    app_transport : ``_SSLProtocolTransport``
        Asynchronous ssl protocol's read-write  transport implementation.
    loop : ``eventThread``
        The respective eventloop of the ssl protocol.
    server_hostname : `None` or `str`
        The ssl protocol's server hostname if applicable.
    server_side : `bool`
        Whether the ssl protocol is server side.
    sslpipe : `None` or ``_SSLPipe``
        Ssl pipe set meanwhile the protocol is connected to feed the ssl data to.
    transport : `None` or `Any`
        Asynchronous transport implementation.
    """
    __slots__ = ('_call_connection_made', '_connection_made_waiter', '_extra', '_handshake_start_time', '_in_handshake',
        '_in_shutdown', '_session_established', '_ssl_context', '_write_backlog', 'app_protocol', 'app_transport',
        'loop', 'server_hostname', 'server_side', 'sslpipe', 'transport',)
    
    def __new__(cls, loop, app_protocol, ssl_context, connection_made_waiter, server_side=False, server_hostname=None,
            call_connection_made=True):
        
        if ssl_context is None:
            ssl_context = _create_transport_context(server_side, server_hostname)
        
        self = object.__new__(cls)
        
        self.server_side = server_side
        if (server_hostname is not None) and server_side:
            server_hostname = None
        
        self.server_hostname = server_hostname
        
        self._ssl_context = ssl_context
        # SSL-specific extra info. More info are set when the handshake
        # completes.
        self._extra = {'sslcontext': ssl_context}

        # App data write buffering
        self._write_backlog = deque()
        
        self._connection_made_waiter = connection_made_waiter
        self.loop = loop
        self.app_protocol = app_protocol
        self.app_transport = _SSLProtocolTransport(self, app_protocol)
        # _SSLPipe instance (None until the connection is made)
        self.sslpipe = None
        self._session_established = False
        self._in_handshake = False
        self._in_shutdown = False
        # transport, ex: SelectorSocketTransport
        self.transport = None
        self._call_connection_made = call_connection_made
        
        return self
    
    def _wakeup_connection_made_waiter(self, exception=None):
        connection_made_waiter = self._connection_made_waiter
        if (connection_made_waiter is not None):
            self._connection_made_waiter = None
            if connection_made_waiter.pending():
                if exception is None:
                    connection_made_waiter.set_result(None)
                else:
                    connection_made_waiter.set_exception(exception)
    
    def connection_made(self, transport):
        self.transport = transport
        self.sslpipe = _SSLPipe(self._ssl_context, self.server_side, self.server_hostname)
        self._start_handshake()
    
    def connection_lost(self, exception):
        if self._session_established:
            self._session_established = False
            app_protocol = self.app_protocol
            self.loop.call_soon(app_protocol.__class__.connection_lost, app_protocol, exception)
        self.transport = None
        self.app_transport = None
        self._wakeup_connection_made_waiter(exception)
    
    def pause_writing(self):
        self.app_protocol.pause_writing()
    
    def resume_writing(self):
        self.app_protocol.resume_writing()
    
    def data_received(self, data):
        try:
            ssldata, appdata = self.sslpipe.feed_ssldata(data)
        except ssl.SSLError:
            self._abort()
            return
        
        for chunk in ssldata:
            self.transport.write(chunk)
        
        for chunk in appdata:
            if chunk:
                self.app_protocol.data_received(chunk)
                continue
            self._start_shutdown()
            break
    
    def eof_received(self):
        try:
            self._wakeup_connection_made_waiter(ConnectionResetError)
            if not self._in_handshake:
                #has no effect whatever it returns when we use ssl
                self.app_protocol.eof_received()
        finally:
            self.transport.close()

    def get_extra_info(self, name, default=None):
        try:
            extra_info = self._extra[name]
        except KeyError:
            transport = self.transport
            if transport is not None:
                extra_info = transport.get_extra_info(name, default)
            else:
                extra_info = default
        
        return extra_info
    
    def _start_shutdown(self):
        if self._in_shutdown:
            return
        self._in_shutdown = True
        self._write_appdata(b'')
    
    def _write_appdata(self, data):
        self._write_backlog.append((data, 0))
        self._process_write_backlog()
    
    def _start_handshake(self):
        self._handshake_start_time = None
        self._in_handshake = True
        #(b'',1) is a special value in _process_write_backlog() to do the SSL handshake
        self._write_backlog.append((b'', 1))
        self.loop.call_soon(self._process_write_backlog)
    
    def _on_handshake_complete(self, handshake_exc):
        self._in_handshake = False
        sslobj = self.sslpipe.ssl_object
        
        try:
            if (handshake_exc is not None):
                raise handshake_exc
            
            peercert = sslobj.getpeercert()
        except BaseException as err:
            self.transport.close()
            if isinstance(err, Exception):
                self._wakeup_connection_made_waiter(err)
                return
            raise
        
        
        # Add extra info that becomes available after handshake.
        extra = self._extra
        extra['peercert'] = peercert
        extra['cipher'] = sslobj.cipher()
        extra['compression']= sslobj.compression()
        extra['ssl_object'] = sslobj

        if self._call_connection_made:
            self.app_protocol.connection_made(self.app_transport)
            
        self._wakeup_connection_made_waiter()
        self._session_established = True
        # In case transport.write() was already called. Don't call
        # immediately _process_write_backlog(), but schedule it:
        # _on_handshake_complete() can be called indirectly from
        # _process_write_backlog(), and _process_write_backlog() is not
        # reentrant.
        self.loop.call_soon(self._process_write_backlog)

    def _process_write_backlog(self):
        # Try to make progress on the write backlog.
        transport = self.transport
        if transport is None:
            return
        
        try:
            for i in range(len(self._write_backlog)):
                data, offset = self._write_backlog[0]
                if data:
                    ssldata, offset = self.sslpipe.feed_appdata(data, offset)
                elif offset:
                    ssldata=self.sslpipe.do_handshake(self._on_handshake_complete)
                    offset = 1
                else:
                    ssldata = self.sslpipe.shutdown(self._finalize)
                    offset = 1

                for chunk in ssldata:
                    transport.write(chunk)

                if offset < len(data):
                    self._write_backlog[0] = (data, offset)
                    # A short write means that a write is blocked on a read
                    # We need to enable reading if it is paused!
                    assert self.sslpipe.need_ssldata
                    if transport._paused:
                        transport.resume_reading()
                    break
                
                # An entire chunk from the backlog was processed. We can
                # delete it and reduce the outstanding buffer size.
                del self._write_backlog[0]
        except BaseException as err:
            if self._in_handshake:
                self._on_handshake_complete(err)
            else:
                self._fatal_error(err, 'Fatal error on SSL transport')
            if not isinstance(err, Exception):
                # BaseException
                raise
    
    def _fatal_error(self, exception,message='Fatal error on transport'):
        # Should be called from exception handler only.
        if not isinstance(exception, (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)):
            self.loop.render_exc_async(exception, [
                message,
                ' exception occured\n',
                repr(self),
                '\n',
                    ])
        
        transport = self.transport
        if (transport is not None):
            transport._force_close(exception)

    def _finalize(self):
        transport = self.transport
        if (transport is not None):
            transport.close()

    def _abort(self):
        transport = self.transport
        if (transport is not None):
            try:
                transport.abort()
            finally:
                self._finalize()


class _SelectorSocketTransport(object):
    __slots__ = ('_extra', 'loop', 'protocol', '_low_water', '_high_water', '_protocol_paused', 'socket', '_sock_fd',
        '_protocol_connected', 'server', 'buffer', '_conn_lost', 'closing', 'eof', 'paused',)
    
    def __init__(self, loop, sock, protocol, waiter=None, extra=None, server=None):
        
        if extra is None:
            extra = {}
        
        self._extra = extra
        self.loop = loop
        self._protocol_paused = False
        self._set_write_buffer_limits()
        
        extra['socket'] = sock
        
        try:
            sockname = sock.getsockname()
        except OSError:
            sockname = None
        extra['sockname'] = sockname
        
        if 'peername' not in extra:
            try:
                peername = sock.getpeername()
            except module_socket.error:
                peername = None
            extra['peername'] = peername
        
        self.socket = sock
        self._sock_fd = sock.fileno()
        
        self.protocol = protocol
        self._protocol_connected = True
        
        self.server = server
        self.buffer = bytearray()
        self._conn_lost = 0  # Set when call to connection_lost scheduled.
        self.closing = False  # Set when close() called.
        if server is not None:
            server._attach()
        
        self.eof = False
        self.paused = False
        
        # Disable the Nagle algorithm -- small writes will be sent without waiting for the TCP ACK.  This generally
        # decreases the latency (in some cases significantly.)
        _set_nodelay(sock)
        
        loop.call_soon(protocol.connection_made, self)
        
        #only start reading when connection_made() has been called
        loop.call_soon(loop.add_reader, self._sock_fd, self._read_ready)
        if (waiter is not None):
            # only wake up the waiter when connection_made() has been called
            loop.call_soon(Future.set_result_if_pending, waiter, None)
    
    def __del__(self):
        socket = self.socket
        if socket is not None:
            socket.close()
    
    def __repr__(self):
        result = [
            '<',
            self.__class__.__name__,
                ]
        
        if self.socket is None:
            result.append(' closed')
        elif self.closing:
            result.append(' closing')
        
        result.append(' fd=')
        result.append(repr(self._sock_fd))
        
        loop = self.loop
        #is the transport open?
        if (loop is not None) and loop.running:
        
            try:
                key = loop.selector.get_key(self._sock_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_READ
            
            result.append(' read=')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)

            try:
                key = loop.selector.get_key(self._sock_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_WRITE

            result.append(' write=<')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)

            result.append(', bufsize=')
            
            bufsize = self.get_write_buffer_size()
            result.append(str(bufsize))
            result.append('>')
        
        result.append('>')
        
        return ''.join(result)
    
    def writelines(self, list_of_data):
        self.write(b''.join(list_of_data))
    
    def get_extra_info(self, name, default=None):
        return self._extra.get(name, default)
    
    def _maybe_pause_protocol(self):
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self._protocol_paused:
            return
        
        self._protocol_paused = True
        try:
            self.protocol.pause_writing()
        except Exception as err:
            self.loop.render_exc_async(err, [
                'Exception occured at:\n',
                repr(self),
                '._maybe_pause_protocol\n',
                    ])
    
    def _maybe_resume_protocol(self):
        if (not self._protocol_paused) or (self.get_write_buffer_size() > self._low_water):
            return
        self._protocol_paused = False
        try:
            self.protocol.resume_writing()
        except Exception as err:
            self.loop.render_exc_async(err, [
                'Exception occured at:\n',
                repr(self),
                '._maybe_resume_protocol\n',
                    ])
    
    def get_write_buffer_limits(self):
        return self._low_water, self._high_water
    
    def _set_write_buffer_limits(self, high=None, low=None):
        if high is None:
            if low is None:
                high = 65536
            else:
                high = low<<2
        
        if low is None:
            low = high>>2
        
        if not high >= low >= 0:
            raise ValueError(f'`high` ({high}) must be `>= low` ({low}) must be `>= 0`.')
        
        self._high_water = high
        self._low_water = low

    def set_write_buffer_limits(self, high=None, low=None):
        self._set_write_buffer_limits(high=high, low=low)
        self._maybe_pause_protocol()
    
    def abort(self):
        self._force_close(None)
    
    def set_protocol(self, protocol):
        self.protocol = protocol
    
    def get_protocol(self):
        return self.protocol
    
    def is_closing(self):
        return self.closing
    
    def close(self):
        if self.closing:
            return
        
        self.closing = True
        self.loop.remove_reader(self._sock_fd)
        if not self.buffer:
            self._conn_lost += 1
            self.loop.remove_writer(self._sock_fd)
            self.loop.call_soon(self._call_connection_lost, None)
    
    def _fatal_error(self, exception, message='Fatal error on transport'):
        if not isinstance(exception, (BrokenPipeError, ConnectionResetError, ConnectionAbortedError)):
            self.loop.render_exc_async(exception, [
                message,
                ' exception occured\n',
                repr(self),
                '\n',
                    ])
        
        self._force_close(exception)
    
    def _force_close(self, exception):
        if self._conn_lost:
            return
        
        if self.buffer:
            self.buffer.clear()
            self.loop.remove_writer(self._sock_fd)
        
        if not self.closing:
            self.closing = True
            self.loop.remove_reader(self._sock_fd)
        
        self._conn_lost += 1
        self.loop.call_soon(self._call_connection_lost, exception)
    
    def _call_connection_lost(self, exception):
        try:
            if self._protocol_connected:
                self.protocol.connection_lost(exception)
        finally:
            self.socket.close()
            self.socket = None
            self.protocol = None
            self.loop = None
            server = self.server
            if (server is not None):
                server._detach()
                self.server = None
    
    def get_write_buffer_size(self):
        return len(self.buffer)
    
    def pause_reading(self):
        if self.closing:
            raise RuntimeError('Cannot pause_reading() when closing')
        if self.paused:
            raise RuntimeError('Already paused')
        self.paused = True
        self.loop.remove_reader(self._sock_fd)
    
    def resume_reading(self):
        if not self.paused:
            raise RuntimeError('Not paused')
        
        self.paused = False
        if self.closing:
            return
        
        self.loop.add_reader(self._sock_fd, self._read_ready)
    
    def _read_ready(self):
        if self._conn_lost:
            return
        try:
            data = self.socket.recv(MAX_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        except Exception as err:
            self._fatal_error(err, 'Fatal read error on socket transport')
        else:
            if data:
                self.protocol.data_received(data)
            elif self.protocol.eof_received():
                # We're keeping the connection open so the
                # protocol can write more, but we still can't
                # receive more, so remove the reader callback.
                self.loop.remove_reader(self._sock_fd)
            else:
                self.close()
    
    def write(self, data):
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f'data argument must be a bytes-like object, got {data.__class__.__name__}')
        if self.eof:
            raise RuntimeError('Cannot call write() after write_eof()')
        if not data:
            return
        
        if self._conn_lost:
            self._conn_lost += 1
            return
        
        if not self.buffer:
            # Optimization: try to send now.
            try:
                n = self.socket.send(data)
            except (BlockingIOError, InterruptedError):
                pass
            except Exception as err:
                self._fatal_error(err, 'Fatal write error on socket transport')
                return
            else:
                data = data[n:]
                if not data:
                    return
            # Not all was written; register write handler.
            self.loop.add_writer(self._sock_fd, self._write_ready)
        
        # Add it to the buffer.
        self.buffer.extend(data)
        self._maybe_pause_protocol()

    def _write_ready(self):
        if self._conn_lost:
            return
        
        try:
            n = self.socket.send(self.buffer)
        except (BlockingIOError, InterruptedError):
            pass
        except Exception as err:
            self.loop.remove_writer(self._sock_fd)
            self.buffer.clear()
            self._fatal_error(err, 'Fatal write error on socket transport')
        else:
            if n:
                del self.buffer[:n]
            
            self._maybe_resume_protocol()  # May append to buffer.
            if not self.buffer:
                self.loop.remove_writer(self._sock_fd)
                
                if self.closing:
                    self._call_connection_lost(None)
                
                elif self.eof:
                    self.socket.shutdown(module_socket.SHUT_WR)
    
    def write_eof(self):
        if self.eof:
            return
        
        self.eof = True
        
        if not self.buffer:
            self.socket.shutdown(module_socket.SHUT_WR)
    
    def can_write_eof(self):
        return True


class _SelectorDatagramTransport(object):
    __slots__ = ('_extra', 'loop', '_protocol_paused', '_high_water', '_low_water', 'protocol', '_sock_fd',
        'buffer', '_conn_lost', 'socket', '_protocol_connected', 'closing', 'address')
    
    def __init__(self, loop, sock, protocol, address=None, waiter=None, extra=None):
        if extra is None:
            extra = {}
        
        self._extra = extra
        
        self.loop = loop
        self._protocol_paused = False
        self._set_write_buffer_limits()
        
        extra['socket'] = sock
        
        try:
            sockname = sock.getsockname()
        except OSError:
            sockname = None
        extra['sockname'] = sockname
        
        if 'peername' not in extra:
            try:
                peername = sock.getpeername()
            except module_socket.error:
                peername = None
            
            extra['peername'] = peername
        
        self.socket = sock
        self._sock_fd = sock.fileno()
        
        self._protocol_connected = False
        self.protocol = protocol
        
        self.buffer = deque()
        self._conn_lost = 0 # Set when call to connection_lost scheduled.
        self.closing = False # Set when close() called.
        
        self.address = address
        self.loop.call_soon(self.protocol.connection_made, self)
        # only start reading when connection_made() has been called
        self.loop.call_soon(self._add_reader, self._sock_fd, self._read_ready)
        
        if waiter is not None:
            # only wake up the waiter when connection_made() has been called
            self.loop.call_soon(Future.set_result_if_pending, waiter, None)
    
    def get_extra_info(self, name, default=None):
        return self._extra.get(name, default)
    
    def is_reading(self):
        raise NotImplementedError
    
    def pause_reading(self):
        raise NotImplementedError
    
    def resume_reading(self):
        raise NotImplementedError
    
    def write(self, data):
        raise NotImplementedError
    
    def writelines(self, list_of_data):
        data = b''.join(list_of_data)
        self.write(data)
    
    def write_eof(self):
        raise NotImplementedError
    
    def can_write_eof(self):
        raise NotImplementedError
    
    def _maybe_pause_protocol(self):
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self._protocol_paused:
            return
            
        self._protocol_paused = True
        try:
            self.protocol.pause_writing()
        except Exception as err:
            self.loop.render_exc_async(err, [
                'Exception occured at:\n',
                repr(self),
                '._maybe_pause_protocol\n',
                    ])
    
    def _maybe_resume_protocol(self):
        if (not self._protocol_paused) or (self.get_write_buffer_size() > self._low_water):
            return
        
        self._protocol_paused = False
        try:
            self.protocol.resume_writing()
        except Exception as err:
            self.loop.render_exc_async(err, [
                'Exception occured at:\n',
                repr(self),
                '._maybe_resume_protocol\n',
                    ])
    
    def get_write_buffer_limits(self):
        return (self._low_water, self._high_water)
    
    def _set_write_buffer_limits(self, high=None, low=None):
        if high is None:
            if low is None:
                high = 65536
            else:
                high = low<<2
        
        if low is None:
            low = high>>2
        
        if not high >= low >= 0:
            raise ValueError(f'`high` ({high}) must be `>= low` ({low}) must be `>= 0`.')
        
        self._high_water = high
        self._low_water = low
    
    def set_write_buffer_limits(self, high=None, low=None):
        self._set_write_buffer_limits(high=high, low=low)
        self._maybe_pause_protocol()
    
    def __repr__(self):
        result = [
            '<',
            self.__class__.__name__,
                ]
        
        if self.socket is None:
            result.append(' closed')
        elif self.closing:
            result.append(' closing')
        
        result.append(' fd=')
        result.append(repr(self._sock_fd))
        
        loop = self.loop
        #is the transport open?
        if (loop is not None) and loop.running:
            try:
                key = loop.selector.get_key(self._sock_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_READ
            
            result.append(' read=')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)
            
            try:
                key = loop.selector.get_key(self._sock_fd)
            except KeyError:
                polling = 0
            else:
                polling = key.events&selectors.EVENT_WRITE
            
            result.append(' write=<')
            if polling:
                state = 'polling'
            else:
                state = 'idle'
            result.append(state)
            
            result.append(', bufsize=')
            
            bufsize = self.get_write_buffer_size()
            result.append(str(bufsize))
            result.append('>')
        
        result.append('>')
        
        return ''.join(result)
    
    def abort(self):
        self._force_close(None)
    
    def set_protocol(self, protocol):
        self.protocol = protocol
        self._protocol_connected = True
    
    def get_protocol(self):
        return self.protocol
    
    def is_closing(self):
        return self.closing
    
    def close(self):
        if self.closing:
            return
        
        self.closing = True
        self.loop.remove_reader(self._sock_fd)
        if not self.buffer:
            self._conn_lost += 1
            self.loop.remove_writer(self._sock_fd)
            self.loop.call_soon(self._call_connection_lost, None)
    
    def __del__(self):
        socket = self.socket
        if socket is not None:
            socket.close()
    
    def _fatal_error(self, exception, message='Fatal error on transport'):
        if not isinstance(exception, OSError):
            self.loop.render_exc_async(exception, [
                message,
                ' exception occured\n',
                repr(self),
                '\n',
                    ])
        
        self._force_close(exception)
    
    def _force_close(self, exception):
        if self._conn_lost:
            return
        
        if self.buffer:
            self.buffer.clear()
            self.loop.remove_writer(self._sock_fd)
        
        if not self.closing:
            self.closing = True
            self.loop.remove_reader(self._sock_fd)
        
        self._conn_lost += 1
        self.loop.call_soon(self._call_connection_lost, exception)
    
    def _call_connection_lost(self, exception):
        try:
            if self._protocol_connected:
                self.protocol.connection_lost(exception)
        finally:
            self.socket.close()
            self.socket = None
            self.protocol = None
            self.loop = None
    
    def _add_reader(self, fd, callback, *args):
        if self.closing:
            return
        
        self.loop.add_reader(fd, callback, *args)
    
    def get_write_buffer_size(self):
        size = 0
        for data, address in self.buffer:
            size += len(data)
        
        return size
    
    def _read_ready(self):
        if self._conn_lost:
            return
        try:
            data, addr = self.socket.recvfrom(MAX_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as err:
            self.protocol.error_received(err)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as err:
            self._fatal_error(err, 'Fatal read error on datagram transport')
        else:
            self.protocol.datagram_received(data, addr)
    
    def sendto(self, data, addr=None):
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError(f'data argument must be a bytes-like object, not {type(data).__name__!r}')
        
        if not data:
            return
        
        address = self.address
        if (address is not None):
            if (addr is not None) or (addr != address):
                raise ValueError(f'Invalid address: must be `None` or `{address!r}`, got {addr!r}.')
            addr = address
        
        if self._conn_lost and (address is not None):
            self._conn_lost += 1
            return
        
        buffer = self.buffer
        if not buffer:
            # Attempt to send it right away first.
            try:
                if self._extra['peername'] is None:
                    self.socket.sendto(data, addr)
                else:
                    self.socket.send(data)
            except (BlockingIOError, InterruptedError):
                self.loop.add_writer(self._sock_fd, self._sendto_ready)
            except OSError as err:
                self.protocol.error_received(err)
                return
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on datagram transport')
                return
            else:
                return
        
        # Ensure that what we buffer is immutable.
        buffer.append((bytes(data), addr))
        self._maybe_pause_protocol()
    
    def _sendto_ready(self):
        buffer = self.buffer
        while buffer:
            data, addr = buffer.popleft()
            try:
                if self._extra['peername'] is None:
                    self.socket.sendto(data, addr)
                else:
                    self.socket.send(data)
            except (BlockingIOError, InterruptedError):
                buffer.appendleft((data, addr))  # Try again later.
                break
            except OSError as err:
                self.protocol.error_received(err)
                return
            except BaseException as err:
                self._fatal_error(err, 'Fatal write error on datagram transport')
                return
        
        self._maybe_resume_protocol() # May append to buffer.
        if not buffer:
            self.loop.remove_writer(self._sock_fd)
            if self.closing:
                self._call_connection_lost(None)



# -*- coding: utf-8 -*-
__all__ = ('Color', )

from math import floor

class Color(int):
    """
    A Color object represents a RGB color. Using int instead of Color is completely fine.
    
    Examples
    --------
    ```py
    >>> # Colors are just int instances, so they can be freely converted between eachother.
    >>> color = Color(0x022587)
    >>> color
    <Color #022587>
    
    >>> # Creating a color from rgb
    >>> color = Color.from_rgb(120, 255, 0)
    >>> color
    <Color #78FF00>
    
    >>> # Or creating a color from a tuple
    >>> color = Color.from_rgb_tuple((100, 20, 255))
    >>> color
    <Color #6414FF>
    
    >>> # Converting back to rgb
    >>> color.as_rgb
    (100, 20, 255)
    
    >>> # Creating color from float color channels
    >>> color = Color.from_rgb_float(1.0, 0.125, 0.862)
    >>> color
    <Color #FF1FDB>
    
    >>> # Converting color back to float color channels
    >>> color.as_rgb_float
    (1.0, 0.12156862745098039, 0.8588235294117647)
    
    >>> # Creating color from html color code
    >>> color = Color.from_html('#ff0000')
    >>> color
    
    >>> # Converting color back to html color code
    <Color #FF0000>
    >>> color.as_html
    '#FF0000'
    
    >>> # html color codes with length of 3 are working as well
    >>> color = Color.from_html('#f00')
    >>> color
    <Color #FF0000>
    ```
    """
    __slots__ = ()
    
    def __repr__(self):
        """Returns the color's representation."""
        return f'<{self.__class__.__name__} #{self:06X}>'
    
    @classmethod
    def from_html(cls, value):
        """
        Converts a HTML color code to a ``Color`` object.
        
        Parameters
        ----------
        value : `str`
            HTML color code.
        
        Returns
        -------
        color : ``Color``
        
        Raises
        ------
        ValueError
            The given value has invalid length or is not hexadecimal.
        """
        value_length = len(value)
        if value_length in (7, 6, 4, 3):
            if value[0] == '#':
                value = value[1:]
                value_length -= 1
            
            if value_length == 6:
                try:
                    color = cls(value, base=16)
                except ValueError:
                    pass
                else:
                    return color
            
            elif value_length == 3:
                try:
                    color_value = int(value, base=16)
                except ValueError:
                    pass
                else:
                    red = (color_value>>8)*17
                    green = ((color_value>>4)&0xf)*17
                    blue = (color_value&0xf)*17
                    return Color((red<<16)|(green<<8)|blue)
        
        raise ValueError('The given value has invalid length or is not hexadecimal.')
    
    @property
    def as_html(self):
        """
        Returns the color in HTML format.
        
        Returns
        -------
        color : `str`
        """
        return f'#{self:06X}'
    
    @classmethod
    def from_rgb_tuple(cls, rgb_tuple):
        """
        Converts a tuple of ints to a ``Color`` object.
        
        Parameters
        ----------
        rgb_tuple : `tuple` (`int`, `int`, `int`)
            A tuple of `3` ints which are in range [0, 255].

        Returns
        -------
        color : ``Color``
        
        Raises
        ------
        ValueError
            - `rgb_tuple` has no length of `3`.
            - Channel value out of expected range.
        """
        if len(rgb_tuple) != 3:
            raise ValueError(f'Rgb tuple should have length `3`, got: {rgb_tuple!r}.')
        
        return cls.from_rgb(*rgb_tuple)
    
    from_tuple = from_rgb_tuple
    
    @property
    def as_rgb_tuple(self):
        """
        Returns the color's red, green and blue channel's value in a tuple of integers.
        
        Returns
        -------
        rgb_tuple : `tuple` (`int`, `int`, `int`)
        """
        return (self>>16, (self>>8)&0x00ff, self&0x0000ff)
    
    as_tuple = as_rgb_tuple
    as_rgb = as_rgb_tuple
    
    @classmethod
    def from_rgb_float_tuple(cls, rgb_tuple):
        """
        Converts a tuple of floats to a ``Color`` object.
        
        Parameters
        ----------
        rgb_tuple : `tuple` (`float`, `float`, `float`)
            A tuple of `3` floats which are in range [0.0, 1.0].
        
        Returns
        -------
        color : ``Color``
        
        Raises
        ------
        ValueError
            - `rgb_tuple` has no length of `3`.
            - Channel value out of expected range.
        """
        if len(rgb_tuple) != 3:
            raise ValueError(f'Rgb tuple should have length `3`, got: {rgb_tuple!r}.')
        
        return cls.from_rgb_float(*rgb_tuple)
    
    from_float_tuple = from_rgb_float_tuple
    
    @property
    def as_rgb_float_tuple(self):
        """
        Returns the color's red, green and blue channel's value as a tuple of floats.
        
        Returns
        -------
        rgb_tuple : `tuple` (`float`, `float`, `float`)
        """
        return ((self>>16)*(1.0/255.0), ((self>>8)&0x00ff)*(1.0/255.0), (self&0x0000ff)*(1.0/255.0))
    
    as_float_tuple = as_rgb_float_tuple
    as_rgb_float = as_rgb_float_tuple
    
    @classmethod
    def from_rgb(cls, red, green, blue):
        """
        Converts red, green and blue channels to a `Color` object.
        
        Does not checks if each element of the tuple is in 0-255, so inputting bad values can yield any error.

        Parameters
        ----------
        red : `int`
            Red channel.
        green : `int`
            Green channel.
        blue : `int`
            Blue channel.

        Returns
        -------
        color : ``Color``
        
        Raises
        ------
            Channel value out of expected range.
        """
        if red > 255 or red < 0:
            raise ValueError(f'Red channel value out of [0, 255] expected range, got {red!r}.')
        if green > 255 or green < 0:
            raise ValueError(f'Green channel value out of [0, 255] expected range, got {green!r}.')
        if blue > 255 or blue < 0:
            raise ValueError(f'Blue channel value out of [0, 255] expected range, got {blue!r}.')
        
        return cls((red<<16)|(green<<8)|blue)
    
    @classmethod
    def from_rgb_float(cls, red, green, blue):
        """
        Converts red, green and blue to a `Color` object.
        
        Does not checks if each element of the tuple is in 0-255, so inputting bad values can yield any error.

        Parameters
        ----------
        red : `float`
            Red channel.
        green : `float`
            Green channel.
        blue : `float`
            Blue channel.

        Returns
        -------
        color : ``Color``
        
        Raises
        ------
        ValueError
            Channel value out of expected range.
        """
        if red > 1.0 or red < 0.0:
            raise ValueError(f'Red channel value out of [0.0, 1.0] expected range, got {red!r}.')
        if green > 1.0 or green < 0.0:
            raise ValueError(f'Green channel value out of [0.0, 1.0] expected range, got {green!r}.')
        if blue > 1.0 or blue < 0.0:
            raise ValueError(f'Blue channel value out of [0.0, 1.0] expected range, got {blue!r}.')
        
        # Convert float values to 8 bit ints.
        red = floor(red*255.0)
        green = floor(green*255.0)
        blue = floor(blue*255.0)
        # Build color.
        return Color((red<<16)|(green<<8)|blue)
    
    @property
    def red(self):
        """
        Returns the color's red channel's value.
        
        Returns
        -------
        red_value : `int`
        """
        return self>>16
    
    r = red
    
    @property
    def green(self):
        """
        Returns the color's green channel's value.
        
        Returns
        -------
        green_value : `int`
        """
        return (self>>8)&0x0000ff
    
    g = green
    
    @property
    def blue(self):
        """
        Returns the color's blue channel's value.
        
        Returns
        -------
        blue_value : `int`
        """
        return self&0x0000ff
    
    b = blue



# -*- coding: utf-8 -*-
from . import checks
from .command import *
from .content_parser import *
from .utils import *
from .client_wrapper_extension import *

__all__ = (
    'checks',
    'setup_ext_commands',
    *client_wrapper_extension.__all__,
    *command.__all__,
    *content_parser.__all__,
    *utils.__all__,
        )

# Add the setup extension function

from ...discord.client import Client
from ...discord.parsers import DEFAULT_EVENT, asynclist, EventWaitforBase

def setup_ext_commands(client, prefix, **kwargs):
    """
    Setups the commands extension of hata on the given client with the given parameters.
    
    Note that this function can be called on a client only once.
    
    The function adds the following attributes to the client:
    - `.command_processer` : ``CommandProcesser``
    - `.commands` : ``_EventHandlerManager``
    
    And the following event handlers are added as well:
    - `message_create` : ``CommandProcesser``
    - `reaction_add` : ``ReactionAddWaitfor`` (Except if other ``EventWaitforBase`` instance is already added.)
    - `reaction_delete` : ``ReactionDeleteWaitfor`` (Except if other ``EventWaitforBase`` instance is already added.)
    
    Parameters
    ----------
    client : ``Client``
        The client on what the extension will be setuped.
    prefix : `str`, (`tuple`, `list`, `deque`) of `str`, or `callable` -> `str`
        The prefix of the client's command processer.
        
        Can be given as `str`, as `tuple`, `list` or `deque` of `str`, or as a `callable`, what accepts `1` argument,
        the respective ``Message`` instance and returns `str`.
    **kwargs : Keyword arguments
        Additional keyword arguments to be passed to the creater ``CommandProcesser``.
    
    Other Parameters
    ----------------
    ignorecase : `bool`
        Whether the prefix should be case insensitive. Defaults to `True`.
    mention_prefix : `bool`
        Whether user mentioning the client as first word in a message's content should be interpretered as a prefix.
        Defaults to `true`
    default_category_name : `None` or `str`
        The command processer's default category's name. Defaults to `None`.
    category_name_rule : `None` or `function`
        Function to generate display names for categories.
        Should accept only 1 argument, what can be `str`  or `None` and should return a `str` instance as well.
    command_name_rule : `None` or `function`
        Function to generate display names for commands.
        Should accept only 1 argument, what is `str` instance and should return a `str` instance as well.
    precheck : `None` or `callable`, Optional
        Function, which desides whether a received message should be processed
        
        The default one filters out every message what's author is a bot account and the channels where the client
        cannot send messages.
        
        The following parameters are passed to it:
        +-----------+---------------+
        | Name      | Type          |
        +===========+===============+
        | client    | ``Client``    |
        +-----------+---------------+
        | message   | ``Message``   |
        +-----------+---------------+
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
        
    Returns
    -------
    command_processer : ``CommandProcesser``
        The created command processer.
    
    Raises
    ------
    TypeError
        If `client` was not given as ``Client`` instance.
    RuntimeError
        - If the given `client` already has `command_processer` or `commands` attribute.
        - If the given `client` has a ``CommandProcesser`` instance added as `message_create` event,
    """
    client_type = client.__class__
    if client_type is not Client:
        raise TypeError(f'Expected type `{Client.__name__}` as client, meanwhile got `{client_type.__name__}`.')
    
    for attr_name in ('command_processer', 'command_processor', 'commands'):
        if hasattr(client, 'attr_name'):
            raise RuntimeError(f'The client already has an attribute named as `{attr_name}`.')
    
    event_message_create = client.events.message_create
    while True:
        if event_message_create is DEFAULT_EVENT:
            break
        
        if type(event_message_create) is asynclist:
            for event in list.__iter__(event_message_create):
                if isinstance(event, CommandProcesser):
                    raise RuntimeError(f'The client already has a `{CommandProcesser.__name__}` instance added as '
                        f'event.')
            break
        
        if isinstance(event_message_create, CommandProcesser):
            raise RuntimeError(f'The client already has a `{CommandProcesser.__name__}` instance added as event.')
        
        break
    
    command_processer = client.events(CommandProcesser(prefix, **kwargs))
    client.command_processer = command_processer
    client.command_processor = command_processer
    client.commands = command_processer.shortcut
    
    event_reaction_add = client.events.reaction_add
    while True:
        if event_reaction_add is DEFAULT_EVENT:
            client.events(ReactionAddWaitfor)
            break
        
        if type(event_reaction_add) is asynclist:
            for event in list.__iter__(event_reaction_add):
                if isinstance(event, EventWaitforBase):
                    break
            else:
                client.events(ReactionAddWaitfor)
            
            break
        
        if isinstance(event_reaction_add, EventWaitforBase):
            break
        
        client.events(ReactionAddWaitfor)
        break
    
    event_reaction_delete = client.events.reaction_delete
    while True:
        if event_reaction_delete is DEFAULT_EVENT:
            client.events(ReactionDeleteWaitfor)
            break
        
        if type(event_reaction_delete) is asynclist:
            for event in list.__iter__(event_reaction_add):
                if isinstance(event, EventWaitforBase):
                    break
            else:
                client.events(ReactionDeleteWaitfor)
            
            break
        
        if isinstance(event_reaction_delete, EventWaitforBase):
            break
        
        client.events(ReactionDeleteWaitfor)
        break
    
    return command_processer


def snapshot_hook():
    from . import snapshot
    

from .. import register_library_extension, add_library_extension_hook
register_library_extension('HuyaneMatsu.commands')
add_library_extension_hook(snapshot_hook, ['HuyaneMatsu.extension_loader'])
del register_library_extension, add_library_extension_hook, snapshot_hook



# -*- coding: utf-8 -*-
import json, re, codecs
from http.cookies import SimpleCookie, CookieError, Morsel
from hashlib import md5, sha1, sha256

try:
    import ssl as module_ssl
except ImportError:
    module_ssl = None

try:
    import cchardet as chardet
except ImportError:
    try:
        import chardet
    except ImportError as err:
        message = ('chardet (or cchardet) is not installed, please make sure it is installed before importing '
            f'{__spec__.parent}')
        
        err.args = (message,)
        err.msg = message
        raise err from None

from .utils import imultidict
from .futures import Task, CancelledError
from .hdrs import METH_POST_ALL, METH_CONNECT, SET_COOKIE, CONTENT_LENGTH, CONNECTION, ACCEPT, ACCEPT_ENCODING, \
    HOST, TRANSFER_ENCODING, COOKIE, CONTENT_ENCODING, AUTHORIZATION, CONTENT_TYPE
from .helpers import BasicAuth
from .multipart import MimeType, create_payload
from .formdata import Formdata
from .protocol import HTTPStreamWriter

json_re = re.compile(r'^application/(?:[\w.+-]+?\+)?json')

class Fingerprint(object):
    __slots__ = ('fingerprint', 'hashfunc',)
    
    HASHFUNC_BY_DIGESTLEN = {
        16: md5,
        20: sha1,
        32: sha256,
            }
    
    def __init__(self, fingerprint):
        digestlen = len(fingerprint)
        try:
            hashfunc = self.HASHFUNC_BY_DIGESTLEN[digestlen]
        except KeyError as err:
            err.args = ('fingerprint has invalid length',)
            raise
        
        if hashfunc is md5 or hashfunc is sha1:
            raise ValueError('md5 and sha1 are insecure and not supported. Use sha256.')
        
        self.hashfunc = hashfunc
        self.fingerprint = fingerprint


    def check(self, transport):
        if not transport.get_extra_info('sslcontext'):
            return
        sslobj = transport.get_extra_info('ssl_object')
        cert = sslobj.getpeercert(binary_form=True)
        got = self.hashfunc(cert).digest()
        if got != self.fingerprint:
            host, port, *_ = transport.get_extra_info('peername')
            raise ValueError(self.fingerprint, got, host, port)

if module_ssl is None:
    SSL_ALLOWED_TYPES = (type(None), )
else:
    SSL_ALLOWED_TYPES = (module_ssl.SSLContext, bool, Fingerprint, type(None))


class ConnectionKey(object):
    """
    Contains information about a host, like proxy, TLS to prevent reusing wrong connections from the pool.
    
    Attributes
    ----------
    host : `str`
        The host's ip address.
    is_ssl : `bool`
        Whether the connection is secure.
    port : `int`
        The host's port.
    proxy_auth : `None` or ``BasicAuth``
        Proxy authorization.
    proxy_url : `None` or ``URL``
        Proxy's url.
    ssl : `None`, ``SSLContext``, `bool`, ``Fingerprint``
        The connection's ssl type.
    """
    __slots__  = ('host', 'is_ssl', 'port', 'proxy_auth',  'proxy_url', 'ssl',) # + 'proxy_header_hash',
    
    def __init__(self, request):
        # proxy_headers = request.proxy_headers
        # if request.proxy_headers is not None:
        #     proxy_header_hash = hash(tuple(proxy_headers.items()))
        # else:
        #     proxy_header_hash = None
        
        self.host = request.host
        self.port = request.port
        self.is_ssl = request.is_ssl()
        self.ssl = request.ssl
        self.proxy_auth = request.proxy_auth
        self.proxy_url = request.proxy_url
        # self.proxy_header_hash = proxy_header_hash
    
    def __repr__(self):
        """Returns the connection key's representation."""
        return f'<{self.__class__.__name__} host={self.host!r}, port={self.port!r}>'
    
    def __eq__(self, other):
        """Returns whether the two connection keys are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.host != other.host:
            return False
        
        if self.port != other.port:
            return False
        
        if self.is_ssl != other.is_ssl:
            return False
        
        if self.ssl is None:
            if other.ssl is not None:
                return False
        else:
            if other.ssl is None:
                return False
            
            if self.ssl != other.ssl:
                return False
            
        if self.proxy_auth is None:
            if other.proxy_auth is not None:
                return False
        else:
            if other.proxy_auth is None:
                return False
            
            if self.proxy_auth != other.proxy_auth:
                return False

        if self.proxy_url is None:
            if other.proxy_url is not None:
                return False
        else:
            if other.proxy_url is None:
                return False
            
            if self.proxy_url != other.proxy_url:
                return False
        
        return True
    
    def __hash__(self):
        """Returns the connection key's hash value."""
        return hash(self.host) ^ (self.port << 17) ^ hash(self.is_ssl) ^ hash(self.ssl) ^ hash(self.proxy_auth) ^ \
               hash(self.proxy_url)

class RequestInfo(object):
    """
    Base information representing a request.
    
    Attributes
    ----------
    headers : ``imultidict``
        The respective request's headers.
    method : `str`
        The repesctive request's method.
    real_url : ``URL``
        The url given to request.
    url : ``URL``
        The requested url without fragments. Can be same as ``.real_url``.
    """
    __slots__ = ('headers', 'method', 'real_url', 'url',)
    def __init__(self, request):
        """
        Creates a new ``RequestInfo`` instance representing the given request.
        
        Parameters
        ----------
        request : ``ClientRequest``
            The represented request.
        """
        self.url = request.url
        self.method = request.method
        self.headers = request.headers
        self.real_url = request.original_url
    
    def __repr__(self):
        """Returns the request info's representation."""
        return f'<{self.__class__.__name__} url={self.url!r}>'


DEFAULT_HEADERS = (
    (ACCEPT, '*/*'),
    (ACCEPT_ENCODING, 'gzip, deflate'),
        )

class ClientRequest(object):
    """
    Http request class used by ``HTTPClient``.
    
    Attributes
    ----------
    auth : `None` or ``BasicAuth``
        Authorization sent with the request.
    body : `None`, ``PayloadBase`` instance
        The request's body.
    chunked : `bool`
        Whether the request is sent chunked.
    compression : `None` or `str`
        Compression used when sending the request.
    headers : `imultidict`
        The headers of the request.
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    method : `str`
        The request's method.
    original_url : ``URL``
        The original url, what was asked to requet.
    proxy_auth : `None` or ``BasicAuth``
        Proxy authorization sent with the request.
    proxy_url : `None` or ``URL``
        Proxy url to use if applicable.
    response : `None` or ``ClientResponse``
        Object representing the received response. Set as `None` till ``.send`` finishes.
    ssl : `None` `None`, ``SSLContext``, `bool`, ``Fingerprint``
        The connection's ssl type.
    url : ``URL``
        The url, what will be requested.
    writer : `None` or ``Task`` of ``.write_bytes``
        Payload writer task, what is present meanwhile the request's payload is sending.
    """
    __slots__ = ('auth', 'body', 'chunked', 'compression', 'headers', 'loop', 'method', 'original_url', 'proxy_auth',
        'proxy_url', 'response', 'ssl', 'url', 'writer',)
    
    def __new__(cls, loop, method, url, headers, data, params, cookies, auth, proxy_url, proxy_auth, ssl):
        """
        Creates a new ``ClientRequest`` instance with the given parameters.
        
        Parameters
        ----------
        loop : ``EventThread``
            The event loop, trough what the request is executed.
        method : `str`
            The request's method.
        url : ``URL``
            The url to request.
        headers : `None`, `dict` or ``imultidict``, Optional
            Headers of the request.
        data : `None`, `bytes-like`, `io-like`, ``FormData`, Optional
            Data to send as the request's body.
        params : `dict` of (`str`, (`str`, `int`, `float`, `bool`)) items
            Query string parameters.
        cookies : `None` or ``CookieJar``
            Cookies OwO.
        auth : `None` or ``BasicAuth``
            Authorization sent with the request.
        proxy_url : `None` or ``URL``
            Proxy url to use if applicable.
        proxy_auth : `None` or ``BasicAuth``
            Proxy authorization sent with the request.
        ssl : `None` `None`, ``SSLContext``, `bool`, ``Fingerprint``
            The connection's ssl type.
        
        Raises
        ------
        TypeError
            - `proxy_auth`'s type is incorrect.
            - ˙Cannot serialize a field of the given `data`.
        ValueError
            - Host could nto be detected from `url`.
            - The `proxy_url`'s scheme is not `http`.
            - `compression` and `Content-Encoding` would be set at the same time.
            - `chunked` cannot be set, because `Transfer-Encoding: chunked` is already set.
            - `chunked` cannot be set, because `Content-Length` header is already present.
        RuntimeError
            - If one of `data`'s field's content has unknown content-encoding.
            - If one of `data`'s field's content has unknown content-transfer-encoding.
        """
        # Convert headers
        headers = imultidict(headers)
        
        # Add extra query parameters to the url and remove fragments
        url = url.extend_query(params)
        request_url = url.with_fragment(None)
        
        if not url.host:
            raise ValueError('Host could not be detected.')
        
        # Check authorization
        if auth is None:
            # If authorization is given, try to detect from url.
            username = url.user
            password = url.password
            
            if (username is not None) and username:
                if password is None:
                    password = ''
                
                auth = BasicAuth(username, password)
        
        # Store auth in headers is applicable.
        if (auth is not None):
            headers[AUTHORIZATION] = auth.encode()
        
        for key, value in DEFAULT_HEADERS:
            headers.setdefault(key, value)
        
        # Add host to headers if not present.
        if HOST not in headers:
            netloc = request_url.raw_host
            if not request_url.is_default_port():
                netloc = f'{netloc}:{request_url.port}'
            
            headers[HOST] = netloc
        
        # Update cookies
        if (cookies is not None) and cookies:
            cookie = SimpleCookie()
            if COOKIE in headers:
                cookie.load(headers.get(COOKIE, ''))
                del headers[COOKIE]
            
            for key, value in cookies.items():
                if isinstance(key, Morsel):
                    # Preserve coded_value
                    try:
                        morsel_value = value.get(value.key)
                    except KeyError:
                        morsel_value = Morsel()
                    
                    morsel_value.set(value.key, value.value, value.coded_value)
                    value = morsel_value
                
                cookie[key] = value
            
            headers[COOKIE] = cookie.output(header='', sep=';').strip()
        
        # Check proxy settings.
        if proxy_url is not None:
            if proxy_auth.scheme != 'http':
                raise ValueError(f'Only http proxies are supported, got {proxy_url!r}.')
            
            if (proxy_auth is not None):
                proxy_auth_type = proxy_auth.__class__
                if proxy_auth_type is not BasicAuth:
                    raise TypeError(f'`proxy_auth` must be `None` or `{BasicAuth.__name__}`, got '
                        f'{proxy_auth_type.__name__}.')
        
        # Needed for transfer data checks
        chunked = True
        compression = headers.get(CONTENT_ENCODING)
        
        # Get request content encoding.
        if (data is not None):
            if data:
                if (compression is not None):
                    if headers.get(CONTENT_ENCODING, ''):
                        raise ValueError('Compression can not be set if `Content-Encoding` header is set.')
                    
                    chunked = True
                
                # formdata
                if isinstance(data, Formdata):
                    data = data()
                else:
                    try:
                        data = create_payload(data, {'disposition': None})
                    except LookupError:
                        data = Formdata.from_fields(data)()
                
                if not chunked:
                    if CONTENT_LENGTH not in headers:
                        size = data.size
                        if size is None:
                            chunked = True
                        else:
                            if CONTENT_LENGTH not in headers:
                                headers[CONTENT_LENGTH] = str(size)
                
                if CONTENT_TYPE not in headers:
                    headers[CONTENT_TYPE] = data.content_type
                
                if data.headers:
                    for key, value in data.headers.items():
                        headers.setdefault(key, value)
            else:
                data = None
        
        # Analyze transfer-encoding header.
        transfer_encoding = headers.get(TRANSFER_ENCODING, '').lower()
        
        if 'chunked' in transfer_encoding:
            if chunked:
                raise ValueError('Chunked can not be set if `Transfer-Encoding: chunked` header is already set.')
        
        elif chunked:
            if CONTENT_LENGTH in headers:
                raise ValueError('Chunked can not be set if `Content-Length` header is set.')
            headers[TRANSFER_ENCODING] = 'chunked'
        
        else:
            if CONTENT_LENGTH not in headers:
                headers[CONTENT_LENGTH] = '0' if data is None else str(len(data))
        
        # Set default content-type.
        if (method in METH_POST_ALL) and (CONTENT_TYPE not in headers):
            headers[CONTENT_TYPE] = 'application/octet-stream'
        
        # Everything seems correct, create the object.
        self = object.__new__(cls)
        
        self.original_url = url
        self.url = request_url
        self.method = method
        self.loop = loop
        self.ssl = ssl
        self.chunked = chunked
        self.compression = compression
        self.body = data
        self.auth = auth
        self.writer = None
        self.response = None
        self.headers = headers
        self.proxy_url = proxy_url
        self.proxy_auth = proxy_auth
        
        return self
    
    def is_ssl(self):
        """
        Returns whether the request is ssl.
        
        Returns
        -------
        is_ssl : `bool`
        """
        return self.url.scheme in ('https', 'wss')
    
    @property
    def connection_key(self):
        """
        Returns the connection key of request.
        
        Returns
        -------
        connection_key : ``ConnectionKey``
        """
        return ConnectionKey(self)
    
    @property
    def request_info(self):
        """
        Returns base information representing the request.
        
        Returns
        -------
        request_info : ``RequestInfo``
        """
        return RequestInfo(self)
    
    @property
    def host(self):
        """
        Returns the request's host.
        
        Returns
        -------
        host : `str`
        """
        return self.url.host
    
    @property
    def port(self):
        """
        Returns the request's port.
        
        Returns
        -------
        port : `int`
        """
        return self.url.port
    
    async def write_bytes(self, writer, connection):
        """
        Writes the request's body..
        
        This method is a coroutine.
        
        Parameters
        ----------
        writer : ``HTTPStreamWriter``
            Writer used to write the request's body into the connection's transport.
        connection : ``Connection``
            Connection of the request with what the payload is sent.
        """
        # Support coroutines that yields bytes objects.
        try:
            body = self.body
            if (body is not None):
                await self.body.write(writer)
            await writer.write_eof()
        except OSError as err:
            new_err = OSError(err.errno, f'Can not write request body for {self.url!r}.')
            new_err.__context__ = err
            new_err.__cause__ = err
            connection.protocol.set_exception(new_err)
        except CancelledError as err:
            if not connection.closed:
               connection.protocol.set_exception(err)
        except BaseException as err:
            connection.protocol.set_exception(err)
            raise
        finally:
            self.writer = None
    
    def send(self, connection):
        """
        Sends the request.
        
        Parameters
        ----------
        connection : ``Connection``
            Connection, what is used to send the request.
        
        Returns
        -------
        response : `coroutine` of ``ClientRespons.start`` ->
        """
        try:
            url = self.url
            if self.method == METH_CONNECT:
                path = f'{url.raw_host}:{url.port}'
            elif (self.proxy_url is not None) and (not self.is_ssl()):
                path = str(url)
            else:
                path = url.raw_path
                if url.raw_query_string:
                    path = f'{path}?{url.raw_query_string}'
            
            protocol = connection.protocol
            writer = HTTPStreamWriter(protocol, self.compression, self.chunked)
            
            protocol.write_http_request(self.method, path, self.headers)
            
            self.writer = Task(self.write_bytes(writer, connection), self.loop)
            
            self.response = response = ClientResponse(self, connection)
            
            return response.start()
        
        except:
            connection.close()
            raise
    
    def terminate(self):
        """
        Terminates the request's writing task if applicable.
        """
        writer = self.writer
        if (writer is not None):
            self.writer = None
            writer.cancel()


class ClientResponse(object):
    """
    Http response class used by ``HTTPClient``.
    
    Attributes
    ----------
    _released : `bool`
        Whether the connection is released.
    body : `None` or `bytes`
        The received response body. Set as `None` if the response body is not yet received, or if it is empty.
    closed : `bool`
        Whether the response is closed.
    connection : `None` or ``Connection``
        Connection used to receive the request response. Set as `None` if the response is ``.close``-d or
        ``.release``-d.
    payload_waiter : `None` or ``Future``
        Future used to retrieve the response's body. It's result is set, when the respective protocol's reader task
        finished.
    cookies : `http.cookies.SimpleCookie`
        Received cookies with the response.
    headers : `None` or ``imultidict``
        Headers of the response. Set when the http response is succesfully received.
    history : `None` or `tuple` of ``ClientResponse``
        Response history. Set as `tuple` of responses from outside.
    loop : ``EventThread``
        The event loop, trough what the request is executed.
    method : `str`
        Method of the respective request.
    status : `None` or `int`
        Received status code. Set as `0` by default.
    url : ``URL``
        The requested url.
    writer : ``Task`` of ``ClientRequest.write_bytes``
        Payload writer task of the respective request.
    raw_message : `None` or ``RawResponseMessage``
        Raw received http response.
    """
    __slots__ = ('_released', 'body', 'closed', 'connection', 'payload_waiter', 'cookies', 'headers', 'history', 'loop',
        'method', 'status', 'url', 'writer', 'raw_message')
       
    def __new__(cls, request, connection):
        """
        Crates a new ``ClientResponse`` instance from the given request and connection.
        
        Parameters
        ----------
        request : ``ClientRequest``
            The respective request.
        connection : ``Connection``
            The connection used to send the request and receive the response.
        """
        self = object.__new__(cls)
        self.loop = request.loop
        self.method = request.method
        self.url = request.original_url
        
        self.writer = request.writer
        self.closed = False
        self.cookies = SimpleCookie()
        self._released = False
        
        self.body = None
        self.status = 0
        self.payload_waiter = None
        self.headers = None
        self.connection = connection
        
        self.raw_message = None
        self.history = None  # will be added later
        
        return self
    
    @property
    def reason(self):
        """
        Returns the server response reason.
        
        reason : `str` or `None`
        """
        message = self.raw_message
        if (message is not None):
            reason = message.reason
            if (reason is not None):
                return reason.decode()
    
    def __del__(self):
        """releases the response if not yet closed."""
        if self.closed:
            return
        
        self._release_connection()
    
    def __repr__(self):
        """Returns the response's representation."""
        ascii_encodable_url = str(self.url)
        
        return f'<{self.__class__.__name__}({ascii_encodable_url}) [{self.status} {self.reason!r}]>'
    
    async def start(self,):
        """
        Starts response processing.
        
        This method is a coroutine.
        
        Returns
        -------
        self : ``ClientResponse``
        """
        try:
            protocol = self.connection.protocol
            
            payload_waiter = protocol.set_payload_reader(protocol._read_http_response())
            self.raw_message = message = await payload_waiter
            protocol.handle_payload_waiter_cancellation()
            payload_reader = protocol.get_payload_reader_task(message)
            if (payload_reader is None):
                payload_waiter = None
                self._response_eof(None)
            else:
                payload_waiter = protocol.set_payload_reader(payload_reader)
                payload_waiter.add_done_callback(self._response_eof)
            
            # response status
            self.status = message.status
            # headers
            self.headers = message.headers
            # OwO
            self.payload_waiter = payload_waiter
            
            # cookies
            for header in self.headers.getall(SET_COOKIE, ()):
                try:
                    self.cookies.load(header)
                except CookieError: # so sad
                    pass
        except:
            self.close()
            raise
        
        return self
    
    def _response_eof(self, future):
        """
        Future callback added to the payload waiter future, to release the used connection.
        
        Parameters
        ----------
        future : ``Future``
            ``.payload_waiter`` future.
        """
        if self.closed:
            return
        
        self.payload_waiter = None
        
        connection = self.connection
        if (connection is not None):
            # Websocket, protocol could be `None`, because connection could be detached.
            if (connection.protocol is not None) and self.raw_message.upgraded:
                return
            
            self._release_connection()
        
        self.closed = True
        self._cleanup_writer()
    
    def _release_connection(self):
        """
        Releases the response's connection.
        
        If the conenction type is "close", closes the protocol as well.
        """
        connection = self.connection
        if connection is None:
            return
        
        headers = self.headers
        if (headers is not None):
            try:
                connection_type = headers[CONNECTION]
            except KeyError:
                pass
            else:
                if connection_type == 'close':
                    protocol = connection.protocol
                    if (protocol is not None):
                        protocol.close()
        
        connection.release()
        self.connection = None
    
    def _notify_content(self):
        """
        Called when response reading is cancelled or released. Sets `ConnectionError` to the respective protocol if
        the payload is still reading.
        """
        payload_waiter = self.payload_waiter
        if (payload_waiter is not None):
            connection = self.connection
            if (connection is not None):
                connection.protocol.set_exception(ConnectionError('Connection closed.'))
        
        self._released = True
    
    def _cleanup_writer(self):
        """
        Cancels the writer task of the respective request. Called when the response is cancelled or released, or if
        reading the whole reponse is done.
        """
        writer = self.writer
        if (writer is not None):
            self.writer = None
            writer.cancel()
    
    async def read(self):
        """
        Reads the response's body.
        
        This method is a coroutine.
        
        Returns
        -------
        body : `bytes`
        """
        payload_waiter = self.payload_waiter
        if (payload_waiter is None):
            body = self.body
        else:
            try:
                self.body = body = await payload_waiter
            finally:
                self.payload_waiter = None
        
        return body
    
    def get_encoding(self):
        """
        Gets the encoding of the response's body.
        
        Returns
        -------
        encoding : `str`
            Defaults to `'utf-8'`.
        """
        ctype = self.headers.get(CONTENT_TYPE, '').lower()
        mimetype = MimeType(ctype)
        
        encoding = mimetype.params.get('charset')
        if encoding is not None:
            try:
                codecs.lookup(encoding)
            except LookupError:
                encoding = None
        
        if encoding is None:
            if mimetype.mtype == 'application' and mimetype.stype == 'json':
                encoding = 'utf-8' # RFC 7159 states that the default encoding is UTF-8.
            else:
                encoding = chardet.detect(self.body)['encoding']
        
        if not encoding:
            encoding = 'utf-8'
        
        return encoding
    
    async def text(self, encoding=None, errors='strict'):
        """
        Loads the response's content as text.
        
        This method is a coroutine.
        
        Parameters
        ----------
        encoding : `None` or `str`, Optional
            If no encoding is given, then detects it from the payload-
        errors : `str`, Optional
            May be given to set a different error handling scheme. The default `errors` value is `'strict'`, meaning
            that encoding errors raise a `UnicodeError`. Other possible values are `'ignore'`, `'replace'`,
            `'xmlcharrefreplace'`, `'backslashreplace'` and any other name registered via `codecs.register_error()`.
        
        Returns
        -------
        text : `str`
        """
        body = await self.read()
        if body is None:
            return
        
        if encoding is None:
            encoding = self.get_encoding()
        
        return body.decode(encoding, errors)
    
    async def json(self, encoding=None, loader=json.loads, content_type='application/json'):
        """
        Loads the response's content as a json.
        
        This method is a coroutine.
        
        Parameters
        ----------
        encoding : None` or `str`, Optionaé
            Encoding to use instead of the response's. If given as `None` (so by default), then will use the response's
            own encoding.
        loader : `callable`, Optional
            Json loader. Defaults to json.loads`.
        content_type : `str`, Optiona
            Content type to use intsead of the default one. Defaults to `'application/json'`.
        
        Returns
        -------
        json : `Any`
        
        Raises
        ------
        TypeError
            If the response's mimetype do not match.
        """
        body = await self.read()
        if body is None:
            return
        
        if content_type is not None:
            received_content_type = self.headers.get(CONTENT_TYPE, '').lower()
            
            if (json_re.match(received_content_type) is None) if (content_type == 'application/json') else \
                    (content_type not in received_content_type):
                raise TypeError(f'Attempt to decode JSON with unexpected mimetype: {received_content_type!r}.')
        
        stripped = body.strip()
        if not stripped:
            return None
        
        if encoding is None:
            encoding = self.get_encoding()
        
        return loader(stripped.decode(encoding))
    
    async def __aenter__(self):
        """
        Enters the client response as an asynchornous context manager.
        
        This method is a coroutine.
        """
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """
        Releases the response if not yet closed.
        
        This method is a coroutine.
        """
        self.release()
        return False

    def close(self):
        """
        Closes the response and it's connection. The used connection will not be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        connection = self.connection
        if (connection is not None):
            self.connection = None
            connection.close()
        
        self._cleanup_writer()
    
    def release(self):
        """
        Releases the response and it's connection. The used connection might be reused after.
        """
        if not self._released:
            self._notify_content()
        
        if self.closed:
            return
        
        self.closed = True
        
        self._release_connection()
        self._cleanup_writer()



# -*- coding: utf-8 -*-
import os, sys, errno
from stat import S_ISCHR, S_ISFIFO, S_ISSOCK
from subprocess import TimeoutExpired, PIPE, Popen
from socket import socketpair as create_socketpair

from .futures import Task, Future, WaitTillAll, future_or_timeout
from .protocol import ReadProtocolBase

IS_AIX = sys.platform.startswith('aix')
LIMIT = 1<<16
MAX_READ_SIZE = 262144

PROCESS_EXIT_DELAY_LIMIT = 10

class UnixReadPipeTransport(object):
    
    __slots__ = ('_extra', 'closing', 'fileno', 'loop', '_paused', 'pipe', 'protocol')
    async def __new__(cls, loop, pipe, protocol, extra=None):
        fileno = pipe.fileno()
        mode = os.fstat(fileno).st_mode
        if not (S_ISFIFO(mode) or S_ISSOCK(mode) or S_ISCHR(mode)):
            raise ValueError('Pipe transport is only for pipes, sockets and character devices.')
        
        self = object.__new__(cls)
        if extra is None:
            extra = {}
        
        extra['pipe'] = pipe
        
        self._extra = extra
        self.loop = loop
        self.pipe = pipe
        self.fileno = fileno
        self.protocol = protocol
        self.closing = False
        self._paused = False
        
        try:
            os.set_blocking(fileno, False)
            # skip 1 callback loop
            future = Future(loop)
            loop.call_soon(Future.set_result_if_pending, future, None)
            await future
            
            protocol.connection_made(self)
            loop.add_reader(fileno, self._read_ready)
        
        except:
            self.close()
            raise
        
        return self
    
    def get_extra_info(self, name, default=None):
        return self._extra.get(name, default)
    
    def __repr__(self):
        return f'<{self.__class__.__name__} fd={self.fileno}>'
    
    def _read_ready(self):
        try:
            data = os.read(self.fileno, MAX_READ_SIZE)
        except (BlockingIOError, InterruptedError):
            pass
        except OSError as err:
            self._fatal_error(err, 'Fatal read error on pipe transport')
        else:
            if data:
                self.protocol.data_received(data)
            else:
                self.closing = True
                loop = self.loop
                loop.remove_reader(self.fileno)
                protocol = self.protocol
                loop.call_soon(protocol.__class__.eof_received, protocol)
                loop.call_soon(self.__class__._call_connection_lost, self, None)
    
    def pause_reading(self):
        if self.closing or self._paused:
            return
        
        self._paused = True
        self.loop.remove_reader(self.fileno)
    
    def resume_reading(self):
        if self.closing or not self._paused:
            return
        
        self._paused = False
        self.loop.add_reader(self.fileno, self._read_ready)
    
    def set_protocol(self, protocol):
        self.protocol = protocol
    
    def get_protocol(self):
        return self.protocol
    
    def is_closing(self):
        return self.closing
    
    def close(self):
        if not self.closing:
            self._close(None)
    
    def __del__(self):
        pipe = self.pipe
        if (pipe is not None):
            pipe.close()
    
    def _fatal_error(self, exception, message='Fatal error on pipe transport'):
        if not (isinstance(exception, OSError) and (exception.errno == errno.EIO)):
            self.loop.render_exc_async(exception, [message, '\non: ', repr(self), '\n'])
        
        self._close(exception)
    
    def _close(self, exception):
        self.closing = True
        loop = self.loop
        loop.remove_reader(self.fileno)
        loop.call_soon(self.__class__._call_connection_lost, self, exception)
    
    def _call_connection_lost(self, exception):
        protocol = self.protocol
        if protocol is None:
            return
        
        try:
            protocol.connection_lost(exception)
        finally:
            pipe = self.pipe
            if (pipe is not None):
                self.pipe = None
                pipe.close()
            
            self.protocol = None

class UnixWritePipeTransport(object):
    
    __slots__ = ('_buffer', '_connection_lost', '_extra', '_high_water', '_low_water', 'closing', 'fileno', 'loop',
        'pipe', 'protocol', 'protocol_paused')
    
    async def __new__(cls, loop, pipe, protocol, extra=None):
        fileno = pipe.fileno()
        mode = os.fstat(fileno).st_mode
        is_char = S_ISCHR(mode)
        is_fifo = S_ISFIFO(mode)
        is_socket = S_ISSOCK(mode)
        if not (is_char or is_fifo or is_socket):
            raise ValueError('Pipe transport is only for pipes, sockets and character devices.')
        
        if extra is None:
            extra = {}
        
        extra['pipe'] = pipe
        
        self = object.__new__(cls)
        self._extra = extra
        self.loop = loop
        self.protocol_paused = False
        self.pipe = pipe
        self.fileno = fileno
        self.protocol = protocol
        self._buffer = bytearray()
        self._connection_lost = 0
        self.closing = False  # Set when close() or write_eof() called.
        
        self._high_water = 65536
        self._low_water = 16384
        
        try:
            os.set_blocking(fileno, False)
            # skip 1 callback loop
            future = Future(loop)
            loop.call_soon(Future.set_result_if_pending, future, None)
            await future
            
            protocol.connection_made(self)
            
            # On AIX, the reader trick (to be notified when the read end of the  socket is closed) only works for
            # sockets. On other platforms it works for pipes and sockets.
            if is_socket or (is_fifo and not IS_AIX):
                loop.add_reader(fileno, self._read_ready)
        except:
            self.close()
            raise
        
        return self
    
    def __repr__(self):
        return f'<{self.__class__.__name__} fd={self.fileno}>'
    
    def get_extra_info(self, name, default=None):
        return self._extra.get(name, default)
    
    def get_write_buffer_size(self):
        return len(self._buffer)
    
    def _read_ready(self):
        # Pipe was closed by peer.
        if self._buffer:
            exception = BrokenPipeError()
        else:
            exception = None
            
        self._close(exception)
    
    def write(self, data):
        if not data:
            return
        
        if isinstance(data, bytearray):
            data = memoryview(data)
        
        if self._connection_lost or self.closing:
            self._connection_lost += 1
            return
        
        buffer = self._buffer
        if not buffer:
            try:
                n = os.write(self.fileno, data)
            except (BlockingIOError, InterruptedError):
                n = 0
            except BaseException as err:
                self._connection_lost += 1
                self._fatal_error(err, 'Fatal write error on pipe transport')
                return
            
            if n == len(data):
                return
            
            if n > 0:
                data = memoryview(data)[n:]
            
            self.loop.add_writer(self.fileno, self._write_ready)
        
        buffer.extend(data)
        self._maybe_pause_protocol()
    
    def _write_ready(self):
        buffer = self._buffer
        
        try:
            n = os.write(self.fileno, buffer)
        except (BlockingIOError, InterruptedError):
            pass
        except BaseException as err:
            buffer.clear()
            self._connection_lost += 1
            self.loop.remove_writer(self.fileno)
            self._fatal_error(err, 'Fatal write error on pipe transport')
        else:
            if n == len(buffer):
                buffer.clear()
                self.loop.remove_writer(self.fileno)
                self._maybe_resume_protocol()  # May append to buffer.
                if self.closing:
                    self.loop.remove_reader(self.fileno)
                    self._call_connection_lost(None)
                return
            
            if n > 0:
                del buffer[:n]
    
    def can_write_eof(self):
        return True
    
    def write_eof(self):
        if self.closing:
            return
        
        self.closing = True
        if not self._buffer:
            loop = self.loop
            loop.remove_reader(self.fileno)
            loop.call_soon(self.__class__._call_connection_lost, self, None)
    
    def set_protocol(self, protocol):
        self.protocol = protocol
    
    def get_protocol(self):
        return self.protocol
    
    def is_closing(self):
        return self.closing
    
    def close(self):
        if (self.pipe is not None) and (not self.closing):
            self.write_eof()
    
    def __del__(self):
        pipe = self.pipe
        if (pipe is not None):
            pipe.close()
    
    def abort(self):
        self._close(None)
    
    def _fatal_error(self, exception, message='Fatal error on pipe transport'):
        if not isinstance(exception, OSError):
            self.loop.render_exc_async(exception, [message, '\non: ', repr(self), '\n'])
        
        self._close(exception)
    
    def _close(self, exception):
        self.closing = True
        
        loop = self.loop
        buffer = self._buffer
        if buffer:
            self.loop.remove_writer(self.fileno)
            buffer.clear()
        
        loop.remove_reader(self.fileno)
        loop.call_soon(self.__class__._call_connection_lost, self, exception)
    
    def _call_connection_lost(self, exception):
        protocol = self.protocol
        if protocol is None:
            return
        
        try:
            protocol.connection_lost(exception)
        finally:
            pipe = self.pipe
            if (pipe is not None):
                self.pipe = None
                pipe.close()
            
            self.protocol = None
    
    def _maybe_pause_protocol(self):
        size = self.get_write_buffer_size()
        if size <= self._high_water:
            return
        
        if self.protocol_paused:
            return
        
        self.protocol_paused = True
        
        protocol = self.protocol
        if protocol is None:
            return
        
        try:
            protocol.pause_writing()
        except BaseException as err:
            self.loop.render_exc_async(err, [
                repr(self), '._maybe_pause_protocol() failed\n'
                'On: ', repr(protocol), '.pause_writing()\n'])
    
    def _maybe_resume_protocol(self):
        if (self.protocol_paused and self.get_write_buffer_size() <= self._low_water):
            self.protocol_paused = False
            protocol = self.protocol
            if (protocol is not None):
                try:
                    protocol.resume_writing()
                except BaseException as err:
                    self.loop.render_exc_async(err, [
                        repr(self), '._maybe_resume_protocol() failed\n'
                        'on: ', repr(protocol), '.resume_writing()\n'])
    
    def get_write_buffer_limits(self):
        return self._low_water, self._high_water
    
    def _set_write_buffer_limits(self, low=None, high=None):
        if high is None:
            if low is None:
                high = 65536
                low = 16384
            else:
                high = low<<2
        else:
            if low is None:
                low = high>>2
        
        if low < 0 or high < low:
            raise ValueError(f'high ({high!r}) must be >= low ({low!r}) must be >= 0')
        
        self._high_water = high
        self._low_water = low
    
    def set_write_buffer_limits(self, low=None, high=None):
        self._set_write_buffer_limits(low=low, high=high)
        self._maybe_pause_protocol()

class SubprocessStreamWriter(object):
    __slots__ = ('loop', 'transport', 'protocol', )
    def __init__(self, loop, transport, protocol):
        self.transport = transport
        self.protocol = protocol
        self.loop = loop
    
    def write(self, data):
        self.transport.write(data)
    
    def writelines(self, data):
        self.transport.writelines(data)
    
    def write_eof(self):
        return self.transport.write_eof()
    
    def can_write_eof(self):
        return self.transport.can_write_eof()
    
    def close(self):
        return self.transport.close()
    
    def is_closing(self):
        return self.transport.is_closing()
    
    async def wait_closed(self):
        await self.protocol._get_close_waiter(self)
    
    def get_extra_info(self, name, default=None):
        return self.transport.get_extra_info(name, default)
    
    async def drain(self):
        if self.transport.is_closing():
            loop = self.loop
            future = Future(loop)
            loop.call_soon(Future.set_result_if_pending, future, None)
            await future
        
        await self.protocol._drain_helper()

class SubprocessWritePipeProtocol(object):
    __slots__ = ('disconnected', 'fd', 'transport', 'process', )
    
    def __init__(self, process, fd):
        self.process = process
        self.fd = fd
        self.transport = None
        self.disconnected = False
    
    def __call__(self):
        return self
    
    def connection_made(self, transport):
        self.transport = transport
    
    def __repr__(self):
        return f'<{self.__class__.__name__} fd={self.fd} pipe={self.transport!r}>'
    
    def connection_lost(self, exception):
        process = self.process
        if (process is not None):
            self.process = None
            process._pipe_connection_lost(self.fd, exception)
            self.disconnected = True
    
    def pause_writing(self):
        self.process._protocol.pause_writing()
    
    def resume_writing(self):
        self.process._protocol.resume_writing()

class SubprocessReadPipeProtocol(SubprocessWritePipeProtocol):
    __slots__ = ()
    def data_received(self, data):
        self.process._pipe_data_received(self.fd, data)
    
    def eof_received(self):
        # connection_lost is received anyways after eof, and connection_lost without exception will cause eof.
        pass

class AsyncProcess(object):
    __slots__ = ('_alive_fds', '_connection_lost', '_drain_waiter', '_exit_waiters', '_extra', '_finished', '_paused',
    '_pending_calls', '_stdin_closed', '_subprocess_stderr_protocol', '_subprocess_stdin_protocol',
    '_subprocess_stdout_protocol', 'closed', 'loop', 'pid', 'process', 'returncode', 'stderr', 'stdin', 'stdout', )
    
    async def __new__(cls, loop, args, shell, stdin, stdout, stderr, bufsize, extra, popen_kwargs):
        if stdin == PIPE:
            # Use a socket pair for stdin, since not all platforms support selecting read events on the write end of a
            # socket (which we use in order to detect closing of the other end).  Notably this is needed on AIX, and
            # works just fine on other platforms.
            stdin_r, stdin_w = create_socketpair()
        else:
            stdin_r = stdin
            stdin_w = None
        
        process = None
        
        try:
            process = Popen(args, shell=shell, stdin=stdin_r, stdout=stdout, stderr=stderr, universal_newlines=False,
                bufsize=bufsize, **popen_kwargs)
            
            if (stdin_w is not None):
                stdin_r.close()
                process.stdin = open(stdin_w.detach(), 'wb', buffering=bufsize)
                stdin_w = None
        except:
            if (process is not None) and (process.poll() is None):
                try:
                    process.kill()
                except ProcessLookupError:
                    pass
            
            raise
        finally:
            if (stdin_w is not None):
                stdin_r.close()
                stdin_w.close()
        
        if extra is None:
            extra = {}
        
        extra['subprocess'] = process
        
        self = object.__new__(cls)
        self._extra = extra
        self.closed = False
        self.loop = loop
        self.process = process
        self.pid = process.pid
        self.returncode = None
        self._exit_waiters = None
        self._pending_calls = []
        self._subprocess_stdin_protocol = None
        self._subprocess_stdout_protocol = None
        self._subprocess_stderr_protocol = None
        self._finished = False
        self.stdin = None
        self.stdout = None
        self.stderr = None
        self._paused = False
        self._drain_waiter = None
        self._connection_lost = False
        self._alive_fds = []
        self._stdin_closed = Future(loop)
        
        try:
            stdin = process.stdin
            if (stdin is not None):
                subprocess_stdin_protocol = SubprocessWritePipeProtocol(self, 0)
                await loop.connect_write_pipe(subprocess_stdin_protocol, stdin)
                self._subprocess_stdin_protocol = subprocess_stdin_protocol
                stdin_transport = subprocess_stdin_protocol.transport
                if (stdin_transport is not None):
                    self.stdin = SubprocessStreamWriter(loop, stdin_transport, protocol=self)
            
            stdout = process.stdout
            if (stdout is not None):
                subprocess_stdout_protocol = SubprocessReadPipeProtocol(self, 1)
                await loop.connect_read_pipe(subprocess_stdout_protocol, stdout)
                self._subprocess_stdout_protocol = subprocess_stdout_protocol
                stdout_transport = subprocess_stdout_protocol.transport
                if (stdout_transport is not None):
                    self.stdout = stdout_protocol = ReadProtocolBase(loop)
                    stdout_protocol.connection_made(stdout_transport)
                    self._alive_fds.append(1)
            
            stderr = process.stderr
            if (stderr is not None):
                subprocess_stderr_protocol = SubprocessReadPipeProtocol(self, 2)
                await loop.connect_read_pipe(subprocess_stderr_protocol, stderr)
                self._subprocess_stderr_protocol = subprocess_stderr_protocol
                stderr_transport = subprocess_stderr_protocol.transport
                if (stderr_transport is not None):
                    self.stderr = ReadProtocolBase(loop)
                    self.stderr.connection_made(stderr_transport)
                    self._alive_fds.append(2)
            
            for pending_call, args in self._pending_calls:
                pending_call(self, *args)
            
            self._pending_calls = None
        except:
            self.close()
            await self.wait()
            raise
        
        return self
    
    def __repr__(self):
        result = [
            '<',
            self.__class__.__name__,
                ]
        
        if self.closed:
            result.append(' closed')
            add_comma = True
        else:
            add_comma = False
        
        stdin = self.stdin
        if (stdin is not None):
            if add_comma:
                result.append(',')
            else:
                add_comma = True
            
            result.append(' stdin=')
            result.append(repr(stdin))
        
        stdout = self.stdout
        if (stdout is not None):
            if add_comma:
                result.append(',')
            else:
                add_comma = True
            
            result.append(' stdout=')
            result.append(repr(stdout))
        
        stderr = self.stderr
        if (stderr is not None):
            if add_comma:
                result.append(',')
            
            result.append(' stderr=')
            result.append(repr(stderr))
        
        result.append('>')
        
        return ''.join(result)
    
    def get_extra_info(self, name, default=None):
        return self._extra.get(name, default)
    
    def is_closing(self):
        return self.closed
    
    def close(self):
        if self.closed:
            return
        
        self.closed = True
        
        pipe_stdin = self._subprocess_stdin_protocol
        if (pipe_stdin is not None):
            pipe_stdin.transport.close()
        
        pipe_stdout = self._subprocess_stdout_protocol
        if (pipe_stdout is not None):
            pipe_stdout.transport.close()
        
        pipe_stderr = self._subprocess_stderr_protocol
        if (pipe_stderr is not None):
            pipe_stderr.transport.close()
        
        process = self.process
        if (process is not None):
            if (self.returncode is None) and (process.poll() is None):
                try:
                    process.kill()
                except ProcessLookupError:
                    pass
        
        self.loop.call_soon(self.__class__._process_exited, self, 0)
    
    __del__ = close
    
    def send_signal(self, signal):
        process = self.process
        if process is None:
            raise ProcessLookupError()
        
        process.send_signal(signal)
    
    def terminate(self):
        process = self.process
        if process is None:
            raise ProcessLookupError()
        
        process.terminate()
    
    async def kill(self):
        process = self.process
        if process is None:
            raise ProcessLookupError()
        
        process.kill()
        
        self.close()
        
        loop = self.loop
        future = Future(loop)
        loop.call_later(0.0, Future.set_result_if_pending, future, None)
        await future
    
    def _pipe_connection_lost(self, fd, exception):
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_pipe_connection_lost(fd, exception)
        else:
            pending_calls.append((self.__class__._do_pipe_connection_lost, (fd, exception)))
        
        self._try_finish()
    
    def _do_pipe_connection_lost(self, fd, exception):
        if fd == 0:
            pipe = self.stdin
            if pipe is not None:
                pipe.close()
            self._do_connection_lost(exception)
            stdin_closed = self._stdin_closed
            if exception is None:
                stdin_closed.set_result_if_pending(None)
            else:
                stdin_closed.set_exception_if_pending(exception)
            return
        
        if fd == 1:
            reader = self.stdout
        elif fd == 2:
            reader = self.stderr
        else:
            reader = None
        
        if (reader is not None):
            reader.connection_lost(exception)
        
        try:
            self._alive_fds.remove(fd)
        except ValueError:
            pass
        
        self._maybe_process_exited()
    
    def _pipe_data_received(self, fd, data):
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._do_pipe_data_received(fd, data)
        else:
            pending_calls.append((self.__class__._do_pipe_data_received, (fd, data)))
    
    def _do_pipe_data_received(self, fd, data):
        if fd == 1:
            reader = self.stdout
        elif fd == 2:
            reader = self.stderr
        else:
            return
        
        if (reader is not None):
            reader.data_received(data)
    
    def _process_exited(self, delayed):
        returncode = self.process.poll()
        if returncode is None:
            if delayed == PROCESS_EXIT_DELAY_LIMIT:
                # do not wait more
                returncode = 255
            else:
                # not yet exited, delay a little bit.
                delayed += 1
                
                self.loop.call_later(0.01*delayed, self.__class__._process_exited, self, delayed)
                return
        
        self.returncode = returncode
        
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._maybe_process_exited()
        else:
            pending_calls.append((self.__class__._maybe_process_exited, ()))
        
        self._try_finish()
        
        # wake up futures waiting for wait()
        exit_waiters = self._exit_waiters
        if (exit_waiters is not None):
            for waiter in self._exit_waiters:
                if not waiter.cancelled():
                    waiter.set_result(returncode)
        
        self._exit_waiters = None
    
    def _maybe_process_exited(self):
        if self._alive_fds:
            return
        
        self.close()
    
    async def wait(self, timeout=None):
        returncode = self.returncode
        if (returncode is not None):
            return returncode
        
        waiter = Future(self.loop)
        
        exit_waiters = self._exit_waiters
        if exit_waiters is None:
            self._exit_waiters = exit_waiters = set()
        
        exit_waiters.add(waiter)
        if (timeout is not None):
            future_or_timeout(waiter, timeout)
        
        try:
            return await waiter
        except TimeoutError:
            try:
                exit_waiters.remove(waiter)
            except ValueError:
                pass
            
            process = self.process
            if process is None:
                args = None
            else:
                args = process.args
            
            raise TimeoutExpired(args, timeout) from None
    
    def _try_finish(self):
        if self.returncode is None:
            return
        
        stdin_protocol = self._subprocess_stdin_protocol
        if (stdin_protocol is None) or stdin_protocol.disconnected:
            return
        
        stdout_protocol = self._subprocess_stdout_protocol
        if (stdout_protocol is None) or stdout_protocol.disconnected:
            return
        
        stderr_protocol = self._subprocess_stderr_protocol
        if (stderr_protocol is None) or stderr_protocol.disconnected:
            return
        
        self._finished = True
        
        pending_calls = self._pending_calls
        if (pending_calls is None):
            self._call_connection_lost(None)
        else:
            pending_calls.append((self.__class__._call_connection_lost, (self, None,)))
    
    def _call_connection_lost(self, exception):
        try:
            self._do_connection_lost(exception)
        finally:
            self.process = None
    
    # Used by .communicate
    async def _feed_stdin(self, input_):
        stdin = self.stdin
        if stdin is None:
            return
        
        stdin.write(input_)
        
        try:
            await stdin.drain()
        except (BrokenPipeError, ConnectionResetError):
            # communicate() ignores BrokenPipeError and ConnectionResetError
            pass
        
        stdin.close()
    
    # Used by .communicate
    async def _read_close_stdout_stream(self):
        stream = self.stdout
        if stream is None:
            return b''
        
        transport = self._subprocess_stdout_protocol.transport
        result = await stream.read()
        transport.close()
        return result
    
    # Used by .communicate
    async def _read_close_stderr_stream(self):
        stream = self.stderr
        if stream is None:
            return b''
        
        transport = self._subprocess_stderr_protocol.transport
        result = await stream.read()
        transport.close()
        return result
    
    async def communicate(self, input_=None, timeout=None):
        tasks = []
        
        loop = self.loop
        if input_ is None:
            stdin_task = None
        else:
            stdin_task = Task(self._feed_stdin(input_), loop)
            tasks.append(stdin_task)
        
        if self.stdout is None:
            stdout_task = None
        else:
            stdout_task = Task(self._read_close_stdout_stream(), loop)
            tasks.append(stdout_task)
        
        if self.stderr is None:
            stderr_task = None
        else:
            stderr_task = Task(self._read_close_stderr_stream(), loop)
            tasks.append(stderr_task)
        
        if tasks:
            future = WaitTillAll(tasks, loop)
            if (timeout is not None):
                future_or_timeout(future, timeout)
            
            _, pending = await future
            
            if pending:
                # timeout occured, cancel the read tasks and raise TimeoutExpired.
                for task in pending:
                    task.cancel()
                
                process = self.process
                if process is None:
                    args = None
                else:
                    args = process.args
                
                raise TimeoutExpired(args, timeout)
            
            if (stdin_task is not None):
                stdin_task.result()
            
            if (stdout_task is None):
                stdout = None
            else:
                stdout = stdout_task.result()
            
            if (stderr_task is None):
                stderr = None
            else:
                stderr = stderr_task.result()
        else:
            stdout = None
            stderr = None
        
        await self.wait()
        
        return stdout, stderr
    
    def poll(self):
        returncode = self.returncode
        if (returncode is not None):
            return returncode
    
    def pause_writing(self):
        self._paused = True
    
    def resume_writing(self):
        self._paused = False
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        drain_waiter.set_result_if_pending(None)
    
    def _do_connection_lost(self, exception):
        self._connection_lost = True
        # Wake up the writer if currently paused.
        if not self._paused:
            return
        
        drain_waiter = self._drain_waiter
        if drain_waiter is None:
            return
        
        self._drain_waiter = None
        if exception is None:
            drain_waiter.set_result_if_pending(None)
        else:
            drain_waiter.set_exception_if_pending(exception)
    
    async def _drain_helper(self):
        if self._connection_lost:
            raise ConnectionResetError('Connection lost')
        
        if not self._paused:
            return
        
        self._drain_waiter = drain_waiter = Future(self)
        await drain_waiter
